#!/usr/bin/env python3
from __future__ import annotations

import argparse
import json
import logging
import math
import multiprocessing
import os
import platform
import queue
import subprocess
import sys
import threading
import time
import uuid
from typing import Any

import dill
import psutil
import zmq
from funcx_common.tasks import TaskState
from parsl.version import VERSION as PARSL_VERSION

from funcx.serialize import FuncXSerializer
from funcx_endpoint.exception_handling import get_error_string, get_result_error_details
from funcx_endpoint.executors.high_throughput.container_sched import naive_scheduler
from funcx_endpoint.executors.high_throughput.mac_safe_queue import mpQueue
from funcx_endpoint.executors.high_throughput.messages import (
    ManagerStatusReport,
    Message,
    Task,
)
from funcx_endpoint.executors.high_throughput.worker_map import WorkerMap
from funcx_endpoint.logging_config import setup_logging

RESULT_TAG = 10
TASK_REQUEST_TAG = 11
HEARTBEAT_CODE = (2**32) - 1


log = logging.getLogger(__name__)


class TaskCancelled(Exception):
    """Task is cancelled by user request."""

    def __init__(self, worker_id, manager_id):
        self.worker_id = worker_id
        self.manager_id = manager_id
        self.tstamp = time.time()

    def __str__(self):
        return (
            "Task cancelled based on user request on manager: "
            f"{self.manager_id}, worker: {self.worker_id}"
        )


class Manager:
    """Manager manages task execution by the workers

                |         0mq              |    Manager         |   Worker Processes
                |                          |                    |
                | <-----Request N task-----+--Count task reqs   |      Request task<--+
    Interchange | -------------------------+->Receive task batch|          |          |
                |                          |  Distribute tasks--+----> Get(block) &   |
                |                          |                    |      Execute task   |
                |                          |                    |          |          |
                | <------------------------+--Return results----+----  Post result    |
                |                          |                    |          |          |
                |                          |                    |          +----------+
                |                          |                IPC-Qeueues

    """

    def __init__(
        self,
        task_q_url="tcp://127.0.0.1:50097",
        result_q_url="tcp://127.0.0.1:50098",
        max_queue_size=10,
        cores_per_worker=1,
        available_accelerators: list[str] | None = None,
        max_workers=float("inf"),
        uid=None,
        heartbeat_threshold=120,
        heartbeat_period=30,
        logdir=None,
        debug=False,
        block_id=None,
        internal_worker_port_range=(50000, 60000),
        worker_mode="singularity_reuse",
        container_cmd_options="",
        scheduler_mode="hard",
        worker_type=None,
        worker_max_idletime=60,
        # TODO : This should be 10ms
        poll_period=100,
    ):
        """
        Parameters
        ----------
        worker_url : str
             Worker url on which workers will attempt to connect back

        uid : str
             string unique identifier

        cores_per_worker : float
             cores to be assigned to each worker. Oversubscription is possible
             by setting cores_per_worker < 1.0. Default=1

        available_accelerators: list of strings
            Accelerators available for workers to use.
            default: empty list

        max_workers : int
             caps the maximum number of workers that can be launched.
             default: infinity

        heartbeat_threshold : int
             Seconds since the last message from the interchange after which the
             interchange is assumed to be un-available, and the manager initiates
             shutdown. Default:120s

             Number of seconds since the last message from the interchange after which
             the worker assumes that the interchange is lost and the manager shuts down.
             Default:120

        heartbeat_period : int
             Number of seconds after which a heartbeat message is sent to the
             interchange

        internal_worker_port_range : tuple(int, int)
             Port range from which the port(s) for the workers to connect to the manager
             is picked.
             Default: (50000,60000)

        worker_mode : str
             Pick between 3 supported modes for the worker:
              1. no_container : Worker launched without containers
              2. singularity_reuse : Worker launched inside a singularity container that
                                     will be reused
              3. singularity_single_use : Each worker and task runs inside a new
                                          container instance.

        container_cmd_options: str
              Container command strings to be added to associated container command.
              For example, singularity exec {container_cmd_options}

        scheduler_mode : str
             Pick between 2 supported modes for the manager:
              1. hard: the manager cannot change the launched container type
              2. soft: the manager can decide whether to launch different containers

        worker_type : str
             If set, the worker type for this manager is fixed. Default: None

        poll_period : int
             Timeout period used by the manager in milliseconds. Default: 10ms
        """
        log.info("Manager started")

        self.context = zmq.Context()
        self.task_incoming = self.context.socket(zmq.DEALER)
        self.task_incoming.setsockopt(zmq.IDENTITY, uid.encode("utf-8"))
        # Linger is set to 0, so that the manager can exit even when there might be
        # messages in the pipe
        self.task_incoming.setsockopt(zmq.LINGER, 0)
        self.task_incoming.connect(task_q_url)

        self.logdir = logdir
        self.debug = debug
        self.block_id = block_id
        self.result_outgoing = self.context.socket(zmq.DEALER)
        self.result_outgoing.setsockopt(zmq.IDENTITY, uid.encode("utf-8"))
        self.result_outgoing.setsockopt(zmq.LINGER, 0)
        self.result_outgoing.connect(result_q_url)

        log.info("Manager connected")

        self.uid = uid

        self.worker_mode = worker_mode
        self.container_cmd_options = container_cmd_options
        self.scheduler_mode = scheduler_mode
        self.worker_type = worker_type
        self.worker_max_idletime = worker_max_idletime
        self.cores_on_node = multiprocessing.cpu_count()
        self.max_workers = max_workers
        self.cores_per_workers = cores_per_worker
        self.available_mem_on_node = round(
            psutil.virtual_memory().available / (2**30), 1
        )
        self.max_worker_count = min(
            max_workers, math.floor(self.cores_on_node / cores_per_worker)
        )

        # Control pinning to accelerators
        self.available_accelerators = available_accelerators or []
        if self.available_accelerators:
            self.max_worker_count = min(
                self.max_worker_count, len(self.available_accelerators)
            )

        self.worker_map = WorkerMap(self.max_worker_count, self.available_accelerators)

        self.internal_worker_port_range = internal_worker_port_range

        self.funcx_task_socket = self.context.socket(zmq.ROUTER)
        self.funcx_task_socket.set_hwm(0)
        self.address = "127.0.0.1"
        self.worker_port = self.funcx_task_socket.bind_to_random_port(
            "tcp://*",
            min_port=self.internal_worker_port_range[0],
            max_port=self.internal_worker_port_range[1],
        )

        log.info(
            "Manager listening on {} port for incoming worker connections".format(
                self.worker_port
            )
        )

        self.task_queues: dict[str, queue.Queue] = {}
        if worker_type:
            self.task_queues[worker_type] = queue.Queue()
        self.outstanding_task_count: dict[str, int] = {}
        self.task_type_mapping: dict[str, str] = {}

        self.pending_result_queue = mpQueue()

        self.max_queue_size = max_queue_size + self.max_worker_count
        self.tasks_per_round = 1

        self.heartbeat_period = heartbeat_period
        self.heartbeat_threshold = heartbeat_threshold
        self.poll_period = poll_period
        self.serializer = FuncXSerializer()
        self.next_worker_q: list[str] = []  # FIFO queue for spinning up workers.
        self.worker_procs: dict[str, subprocess.Popen] = {}

        self.task_status_deltas: dict[str, tuple[float, TaskState]] = {}

        self._kill_event = threading.Event()
        self._result_pusher_thread = threading.Thread(
            target=self.push_results, args=(self._kill_event,), name="Result-Pusher"
        )
        self._status_report_thread = threading.Thread(
            target=self._status_report_loop,
            args=(self._kill_event,),
            name="Status-Report",
        )
        self.container_switch_count = 0

        self.poller = zmq.Poller()
        self.poller.register(self.task_incoming, zmq.POLLIN)
        self.poller.register(self.funcx_task_socket, zmq.POLLIN)
        self.task_worker_map: dict[str, Any] = {}

        self.task_done_counter = 0
        self.task_finalization_lock = threading.Lock()

    def create_reg_message(self):
        """Creates a registration message to identify the worker to the interchange"""
        msg = {
            "parsl_v": PARSL_VERSION,
            "python_v": "{}.{}.{}".format(
                sys.version_info.major, sys.version_info.minor, sys.version_info.micro
            ),
            "max_worker_count": self.max_worker_count,
            "cores": self.cores_on_node,
            "mem": self.available_mem_on_node,
            "block_id": self.block_id,
            "worker_type": self.worker_type,
            "os": platform.system(),
            "hname": platform.node(),
            "dir": os.getcwd(),
        }
        b_msg = json.dumps(msg).encode("utf-8")
        return b_msg

    def pull_tasks(self, kill_event):
        """Pull tasks from the incoming tasks 0mq pipe onto the internal
        pending task queue


        While :
            receive results and task requests from the workers
            receive tasks/heartbeats from the Interchange
            match tasks to workers
            if task doesn't have appropriate worker type:
                 launch worker of type.. with LRU or some sort of caching strategy.
            if workers >> tasks:
                 advertize available capacity

        Parameters:
        -----------
        kill_event : threading.Event
              Event to let the thread know when it is time to die.
        """
        log.info("starting")

        # Send a registration message
        msg = self.create_reg_message()
        log.debug(f"Sending registration message: {msg}")
        self.task_incoming.send(msg)
        last_interchange_contact = time.time()
        task_recv_counter = 0

        poll_timer = self.poll_period

        new_worker_map = None
        while not kill_event.is_set():
            # Disabling the check on ready_worker_queue disables batching
            log.trace("Loop start")
            pending_task_count = task_recv_counter - self.task_done_counter
            ready_worker_count = self.worker_map.ready_worker_count()
            log.trace(
                "pending_task_count: %s, Ready_worker_count: %s",
                pending_task_count,
                ready_worker_count,
            )

            if pending_task_count < self.max_queue_size and ready_worker_count > 0:
                ads = self.worker_map.advertisement()
                log.trace(f"Requesting tasks: {ads}")
                msg = dill.dumps(ads)
                self.task_incoming.send(msg)

            # Receive results from the workers, if any
            socks = dict(self.poller.poll(timeout=poll_timer))

            if (
                self.funcx_task_socket in socks
                and socks[self.funcx_task_socket] == zmq.POLLIN
            ):
                self.poll_funcx_task_socket()

            # Receive task batches from Interchange and forward to workers
            if self.task_incoming in socks and socks[self.task_incoming] == zmq.POLLIN:

                # If we want to wrap the task_incoming polling into a separate function,
                # we need to
                #   self.poll_task_incoming(
                #       poll_timer,
                #       last_interchange_contact,
                #       kill_event,
                #       task_revc_counter
                #   )
                poll_timer = 0
                _, pkl_msg = self.task_incoming.recv_multipart()
                message = dill.loads(pkl_msg)
                last_interchange_contact = time.time()

                if message == "STOP":
                    log.critical("Received stop request")
                    kill_event.set()
                    break

                elif type(message) == tuple and message[0] == "TASK_CANCEL":
                    with self.task_finalization_lock:
                        task_id = message[1]
                        log.info(f"Received TASK_CANCEL request for task: {task_id}")
                        if task_id not in self.task_worker_map:
                            log.warning(f"Task:{task_id} is not in task_worker_map.")
                            log.warning("Possible duplicate cancel or race-condition")
                            continue
                        # Cancel task by killing the worker it is on
                        worker_id_raw = self.task_worker_map[task_id]["worker_id"]
                        worker_to_kill = self.task_worker_map[task_id][
                            "worker_id"
                        ].decode("utf-8")
                        worker_type = self.task_worker_map[task_id]["task_type"]
                        log.debug(
                            "Cancelling task running on worker: %s",
                            self.task_worker_map[task_id],
                        )
                        try:
                            log.info(f"Removing worker:{worker_id_raw} from map")
                            self.worker_map.start_remove_worker(worker_type)

                            log.info(
                                f"Popping worker:{worker_to_kill} from worker_procs"
                            )
                            proc = self.worker_procs.pop(worker_to_kill)
                            log.warning(f"Sending process:{proc.pid} terminate signal")
                            proc.terminate()
                            try:
                                proc.wait(1)  # Wait 1 second before attempting SIGKILL
                            except subprocess.TimeoutExpired:
                                log.exception("Process did not terminate in 1 second")
                                log.warning(f"Sending process:{proc.pid} kill signal")
                                proc.kill()
                            else:
                                log.debug(
                                    f"Worker process exited with : {proc.returncode}"
                                )

                            # Now that the worker is dead, remove it from worker map
                            self.worker_map.remove_worker(worker_id_raw)
                            raise TaskCancelled(worker_to_kill, self.uid)
                        except Exception as e:
                            log.exception(f"Raise exception, handling: {e}")
                            result_package = {
                                "task_id": task_id,
                                "container_id": worker_type,
                                "error_details": get_result_error_details(e),
                                "exception": get_error_string(tb_levels=0),
                            }
                            self.pending_result_queue.put(dill.dumps(result_package))

                        worker_proc = self.worker_map.add_worker(
                            worker_id=str(self.worker_map.worker_id_counter),
                            worker_type=self.worker_type,
                            container_cmd_options=self.container_cmd_options,
                            address=self.address,
                            debug=self.debug,
                            uid=self.uid,
                            logdir=self.logdir,
                            worker_port=self.worker_port,
                        )
                        self.worker_procs.update(worker_proc)
                        self.task_worker_map.pop(task_id)
                        self.remove_task(task_id)

                elif message == HEARTBEAT_CODE:
                    log.debug("Got heartbeat from interchange")

                else:
                    tasks = [
                        (rt["local_container"], Message.unpack(rt["raw_buffer"]))
                        for rt in message
                    ]

                    task_recv_counter += len(tasks)
                    log.debug(
                        "Got tasks: {} of {}".format(
                            [t[1].task_id for t in tasks], task_recv_counter
                        )
                    )

                    for task_type, task in tasks:
                        log.debug(f"Task is of type: {task_type}")

                        if task_type not in self.task_queues:
                            self.task_queues[task_type] = queue.Queue()
                        if task_type not in self.outstanding_task_count:
                            self.outstanding_task_count[task_type] = 0
                        self.task_queues[task_type].put(task)
                        self.outstanding_task_count[task_type] += 1
                        self.task_type_mapping[task.task_id] = task_type
                        log.debug(
                            "Got task: Outstanding task counts: {}".format(
                                self.outstanding_task_count
                            )
                        )
                        log.debug(
                            f"Task {task.task_id} pushed to task queue "
                            f"for type: {task_type}"
                        )

            else:
                log.trace("No incoming tasks")
                # Limit poll duration to heartbeat_period
                # heartbeat_period is in s vs poll_timer in ms
                if not poll_timer:
                    poll_timer = self.poll_period
                poll_timer = min(self.heartbeat_period * 1000, poll_timer * 2)

                # Only check if no messages were received.
                if time.time() > last_interchange_contact + self.heartbeat_threshold:
                    log.critical(
                        "Missing contact with interchange beyond heartbeat_threshold"
                    )
                    kill_event.set()
                    log.critical("Killing all workers")
                    for proc in self.worker_procs.values():
                        proc.kill()
                    log.critical("Exiting")
                    break

            log.trace(f"To-Die Counts: {self.worker_map.to_die_count}")
            log.trace(
                "Alive worker counts: {}".format(
                    self.worker_map.total_worker_type_counts
                )
            )

            new_worker_map = naive_scheduler(
                self.task_queues,
                self.outstanding_task_count,
                self.max_worker_count,
                new_worker_map,
                self.worker_map.to_die_count,
            )
            log.trace(f"New worker map: {new_worker_map}")

            # NOTE: Wipes the queue -- previous scheduling loops don't affect what's
            # needed now.
            self.next_worker_q, need_more = self.worker_map.get_next_worker_q(
                new_worker_map
            )

            # Spin up any new workers according to the worker queue.
            # Returns the total number of containers that have spun up.
            self.worker_procs.update(
                self.worker_map.spin_up_workers(
                    self.next_worker_q,
                    mode=self.worker_mode,
                    debug=self.debug,
                    container_cmd_options=self.container_cmd_options,
                    address=self.address,
                    uid=self.uid,
                    logdir=self.logdir,
                    worker_port=self.worker_port,
                )
            )
            log.trace(f"Worker processes: {self.worker_procs}")

            #  Count the workers of each type that need to be removed
            spin_downs, container_switch_count = self.worker_map.spin_down_workers(
                new_worker_map,
                worker_max_idletime=self.worker_max_idletime,
                need_more=need_more,
                scheduler_mode=self.scheduler_mode,
            )
            self.container_switch_count += container_switch_count
            log.trace(
                "Container switch count: total {}, cur {}".format(
                    self.container_switch_count, container_switch_count
                )
            )

            for w_type in spin_downs:
                self.remove_worker_init(w_type)

            current_worker_map = self.worker_map.get_worker_counts()
            for task_type in current_worker_map:
                if task_type == "unused":
                    continue

                # *** Match tasks to workers *** #
                else:
                    available_workers = current_worker_map[task_type]
                    log.trace(
                        "Available workers of type {}: {}".format(
                            task_type, available_workers
                        )
                    )

                    for _i in range(available_workers):
                        if (
                            task_type in self.task_queues
                            and not self.task_queues[task_type].qsize() == 0
                            and not self.worker_map.worker_queues[task_type].qsize()
                            == 0
                        ):

                            log.debug(
                                "Task type {} has task queue size {}".format(
                                    task_type, self.task_queues[task_type].qsize()
                                )
                            )
                            log.debug(
                                "... and available workers: {}".format(
                                    self.worker_map.worker_queues[task_type].qsize()
                                )
                            )

                            self.send_task_to_worker(task_type)

    def poll_funcx_task_socket(self, test=False):
        try:
            w_id, m_type, message = self.funcx_task_socket.recv_multipart()
            if m_type == b"REGISTER":
                reg_info = dill.loads(message)
                log.debug(f"Registration received from worker:{w_id} {reg_info}")
                self.worker_map.register_worker(w_id, reg_info["worker_type"])

            elif m_type == b"TASK_RET":
                # the following steps are also shared by task_cancel
                with self.task_finalization_lock:
                    log.debug(f"Result received from worker: {w_id}")
                    task_id = dill.loads(message)["task_id"]
                    try:
                        self.remove_task(task_id)
                    except KeyError:
                        log.exception(f"Task:{task_id} missing in task structure")
                    else:
                        self.pending_result_queue.put(message)
                        self.worker_map.put_worker(w_id)

            elif m_type == b"WRKR_DIE":
                log.debug(f"[WORKER_REMOVE] Removing worker {w_id} from worker_map...")
                log.debug(
                    "Ready worker counts: {}".format(
                        self.worker_map.ready_worker_type_counts
                    )
                )
                log.debug(
                    "Total worker counts: {}".format(
                        self.worker_map.total_worker_type_counts
                    )
                )
                self.worker_map.remove_worker(w_id)
                proc = self.worker_procs.pop(w_id.decode())
                if not proc.poll():
                    try:
                        proc.wait(timeout=1)
                    except subprocess.TimeoutExpired:
                        log.warning(
                            "[WORKER_REMOVE] Timeout waiting for worker %s process to "
                            "terminate",
                            w_id,
                        )
                log.debug(f"[WORKER_REMOVE] Removing worker {w_id} process object")
                log.debug(f"[WORKER_REMOVE] Worker processes: {self.worker_procs}")

            if test:
                return dill.loads(message)

        except Exception:
            log.exception("Unhandled exception while processing worker messages")

    def remove_task(self, task_id: str):
        task_type = self.task_type_mapping.pop(task_id)
        self.task_status_deltas.pop(task_id, None)
        self.outstanding_task_count[task_type] -= 1
        self.task_done_counter += 1

    def send_task_to_worker(self, task_type):
        task = self.task_queues[task_type].get()
        worker_id = self.worker_map.get_worker(task_type)

        log.debug(f"Sending task {task.task_id} to {worker_id}")
        # TODO: Some duplication of work could be avoided here
        to_send = [
            worker_id,
            dill.dumps(task.task_id),
            dill.dumps(task.container_id),
            task.pack(),
        ]
        self.funcx_task_socket.send_multipart(to_send)
        self.worker_map.update_worker_idle(task_type)
        if task.task_id != "KILL":
            log.debug(f"Set task {task.task_id} to RUNNING")
            ts = (time.monotonic(), TaskState.RUNNING)
            self.task_status_deltas[task.task_id] = ts
            self.task_worker_map[task.task_id] = {
                "worker_id": worker_id,
                "task_type": task_type,
            }
        log.debug("Sending complete")

    def _status_report_loop(self, kill_event: threading.Event):
        log.debug("Manager status reporting loop starting")

        while not kill_event.wait(timeout=self.heartbeat_period):
            msg = ManagerStatusReport(
                self.task_status_deltas,
                self.container_switch_count,
            )
            log.info(f"Sending status report to interchange: {msg.task_statuses}")
            self.pending_result_queue.put(msg)
            log.info("Clearing task deltas")
            self.task_status_deltas.clear()

    def push_results(self, kill_event, max_result_batch_size=1):
        """Listens on the pending_result_queue and sends out results via 0mq

        Parameters:
        -----------
        kill_event : threading.Event
              Event to let the thread know when it is time to die.
        """

        log.debug("Starting thread")

        push_poll_period = (
            max(10, self.poll_period) / 1000
        )  # push_poll_period must be atleast 10 ms
        log.debug(f"push poll period: {push_poll_period}")

        last_beat = time.time()
        items = []

        while not kill_event.is_set():
            try:
                r = self.pending_result_queue.get(block=True, timeout=push_poll_period)
                # This avoids the interchange searching and attempting to unpack every
                # message in case it's a status report.
                # (It would be better to use Task Messages eventually to make this more
                #  uniform)
                # TODO: use task messages, and don't have to prepend
                if isinstance(r, ManagerStatusReport):
                    items.insert(0, r.pack())
                else:
                    items.append(r)
            except queue.Empty:
                pass
            except Exception as e:
                log.exception(f"Got an exception: {e}")

            # If we have reached poll_period duration or timer has expired, we send
            # results
            if (
                len(items) >= self.max_queue_size
                or time.time() > last_beat + push_poll_period
            ):
                last_beat = time.time()
                if items:
                    self.result_outgoing.send_multipart(items)
                    items = []

        log.critical("Exiting")

    def remove_worker_init(self, worker_type):
        """
        Kill/Remove a worker of a given worker_type.

        Add a kill message to the task_type queue.

        Assumption : All workers of the same type are uniform, and therefore don't
                     discriminate when killing.
        """

        log.debug(
            "[WORKER_REMOVE] Appending KILL message to worker queue {}".format(
                worker_type
            )
        )
        self.worker_map.start_remove_worker(worker_type)
        task = Task(task_id="KILL", container_id="RAW", task_buffer="KILL")
        self.task_queues[worker_type].put(task)

    def start(self):
        """
        * while True:
            Receive tasks and start appropriate workers
            Push tasks to available workers
            Forward results
        """

        if self.worker_type and self.scheduler_mode == "hard":
            log.debug(
                "[MANAGER] Start an initial worker with worker type {}".format(
                    self.worker_type
                )
            )
            self.worker_procs.update(
                self.worker_map.add_worker(
                    worker_id=str(self.worker_map.worker_id_counter),
                    worker_type=self.worker_type,
                    container_cmd_options=self.container_cmd_options,
                    address=self.address,
                    debug=self.debug,
                    uid=self.uid,
                    logdir=self.logdir,
                    worker_port=self.worker_port,
                )
            )

        log.debug("Initial workers launched")
        self._result_pusher_thread.start()
        self._status_report_thread.start()
        self.pull_tasks(self._kill_event)
        log.info("Waiting")


def cli_run():

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-d", "--debug", action="store_true", help="Count of apps to launch"
    )
    parser.add_argument(
        "-l",
        "--logdir",
        default="process_worker_pool_logs",
        help="Process worker pool log directory",
    )
    parser.add_argument(
        "-u",
        "--uid",
        default=str(uuid.uuid4()).split("-")[-1],
        help="Unique identifier string for Manager",
    )
    parser.add_argument(
        "-b", "--block_id", default=None, help="Block identifier string for Manager"
    )
    parser.add_argument(
        "-c",
        "--cores_per_worker",
        default="1.0",
        help="Number of cores assigned to each worker process. Default=1.0",
    )
    parser.add_argument(
        "-a",
        "--available-accelerators",
        default=(),
        nargs="*",
        help="List of available accelerators",
    )
    parser.add_argument(
        "-t", "--task_url", required=True, help="REQUIRED: ZMQ url for receiving tasks"
    )
    parser.add_argument(
        "--max_workers",
        default=float("inf"),
        help="Caps the maximum workers that can be launched, default:infinity",
    )
    parser.add_argument(
        "--hb_period",
        default=30,
        help="Heartbeat period in seconds. Uses manager default unless set",
    )
    parser.add_argument(
        "--hb_threshold",
        default=120,
        help="Heartbeat threshold in seconds. Uses manager default unless set",
    )
    parser.add_argument("--poll", default=10, help="Poll period used in milliseconds")
    parser.add_argument(
        "--worker_type", default=None, help="Fixed worker type of manager"
    )
    parser.add_argument(
        "--worker_mode",
        default="singularity_reuse",
        help=(
            "Choose the mode of operation from "
            "(no_container, singularity_reuse, singularity_single_use"
        ),
    )
    parser.add_argument(
        "--container_cmd_options",
        default="",
        help=("Container cmd options to add to container startup cmd"),
    )
    parser.add_argument(
        "--scheduler_mode",
        default="soft",
        help=("Choose the mode of scheduler (hard, soft"),
    )
    parser.add_argument(
        "-r",
        "--result_url",
        required=True,
        help="REQUIRED: ZMQ url for posting results",
    )

    args = parser.parse_args()

    setup_logging(
        logfile=os.path.join(args.logdir, args.uid, "manager.log"), debug=args.debug
    )

    try:
        log.info(f"Python version: {sys.version}")
        log.info(f"Debug logging: {args.debug}")
        log.info(f"Log dir: {args.logdir}")
        log.info(f"Manager ID: {args.uid}")
        log.info(f"Block ID: {args.block_id}")
        log.info(f"cores_per_worker: {args.cores_per_worker}")
        log.info(f"available_accelerators: {args.available_accelerators}")
        log.info(f"task_url: {args.task_url}")
        log.info(f"result_url: {args.result_url}")
        log.info(f"hb_period: {args.hb_period}")
        log.info(f"hb_threshold: {args.hb_threshold}")
        log.info(f"max_workers: {args.max_workers}")
        log.info(f"poll_period: {args.poll}")
        log.info(f"worker_mode: {args.worker_mode}")
        log.info(f"container_cmd_options: {args.container_cmd_options}")
        log.info(f"scheduler_mode: {args.scheduler_mode}")
        log.info(f"worker_type: {args.worker_type}")

        manager = Manager(
            task_q_url=args.task_url,
            result_q_url=args.result_url,
            uid=args.uid,
            block_id=args.block_id,
            cores_per_worker=float(args.cores_per_worker),
            available_accelerators=args.available_accelerators,
            max_workers=args.max_workers
            if args.max_workers == float("inf")
            else int(args.max_workers),
            heartbeat_threshold=int(args.hb_threshold),
            heartbeat_period=int(args.hb_period),
            logdir=args.logdir,
            debug=args.debug,
            worker_mode=args.worker_mode,
            container_cmd_options=args.container_cmd_options,
            scheduler_mode=args.scheduler_mode,
            worker_type=args.worker_type,
            poll_period=int(args.poll),
        )
        manager.start()

    except Exception as e:
        log.critical("process_worker_pool exiting from an exception")
        log.exception(f"Caught error: {e}")
        raise
    else:
        log.info("process_worker_pool main event loop exiting normally")
        print("PROCESS_WORKER_POOL main event loop exiting normally")


if __name__ == "__main__":
    cli_run()



"""
This module contains logging configuration for the funcx-endpoint application.
"""

import logging
import logging.config
import logging.handlers
import os
import re
import sys
import typing as t

log = logging.getLogger(__name__)

DEFAULT_FORMAT = (
    "%(created)f %(asctime)s %(levelname)s %(processName)s-%(process)d "
    "%(threadName)s-%(thread)d %(name)s:%(lineno)d %(funcName)s "
    "%(message)s"
)

_ital = "\033[3m"
_redb = "\033[41m"
_teal = "\033[32m"
_yel = "\033[33m"
_byel = "\033[93m"
_yelb = "\033[43m"
_purp = "\033[35m"
_cyan = "\033[36m"
_gray = "\033[37m"
_grayonb = "\033[37;40m"
_r = "\033[m"
_C_BASE = (
    f"{_teal}%(created)f{_r} {_yel}%(asctime)s{_r} {_ital}%(levelname)s{_r}"
    " %(processName)s-%(process)d %(threadName)s-%(thread)d"
    f" %(name)s:{_cyan}%(lineno)d{_r} {_purp}%(funcName)s{_r}"
)
COLOR_ERROR = _redb
COLOR_WARNING = _yelb
COLOR_INFO = _gray
COLOR_DEBUG = _grayonb
C_ERROR_FMT = _C_BASE + f" {COLOR_ERROR}%(message)s{_r}"
C_WARNING_FMT = _C_BASE + f" {COLOR_WARNING}%(message)s{_r}"
C_INFO_FMT = _C_BASE + f" {COLOR_INFO}%(message)s{_r}"
C_DEBUG_FMT = _C_BASE + f" {COLOR_DEBUG}%(message)s{_r}"


class FuncxConsoleFormatter(logging.Formatter):
    """
    For internal use only.
    This formatter handles output to standard streams in the following way:

    if 'debug' is False (default):
        info messages and below are treated as "user output" and are minimally decorated
        warning messages and up are given a full format, so are debug messages

    if 'debug' is True:
        all messages are given the full format
    """

    _u = "[0-9A-Fa-f]"  # convenience
    _uuid_re = f"{_u}{{8}}-{_u}{{4}}-{_u}{{4}}-{_u}{{4}}-{_u}{{12}}"
    # match uuids for colorization that have not otherwise already been colorized
    uuid_re = re.compile(rf"(?<!\dm)({_uuid_re})")

    def __init__(
        self,
        debug: bool = False,
        no_color: bool = False,
        fmt: str = "",
        datefmt: str = "%Y-%m-%d %H:%M:%S",
    ) -> None:
        super().__init__()

        self.use_color = debug and not no_color and sys.stderr.isatty()

        if fmt:
            d_fmt, i_fmt, w_fmt, e_fmt = fmt, fmt, fmt, fmt
        else:
            d_fmt, i_fmt, w_fmt, e_fmt = (
                C_DEBUG_FMT,
                C_INFO_FMT,
                C_WARNING_FMT,
                C_ERROR_FMT,
            )

        if not self.use_color:
            ansi_re = re.compile("\033.*?m")
            d_fmt = ansi_re.sub("", d_fmt)
            i_fmt = ansi_re.sub("", i_fmt)
            w_fmt = ansi_re.sub("", w_fmt)
            e_fmt = ansi_re.sub("", e_fmt)

        if debug:
            self._error_formatter = logging.Formatter(fmt=e_fmt, datefmt=datefmt)
            self._warning_formatter = logging.Formatter(fmt=w_fmt, datefmt=datefmt)
            self._debug_formatter = logging.Formatter(fmt=d_fmt, datefmt=datefmt)
            self._info_formatter = logging.Formatter(fmt=i_fmt, datefmt=datefmt)
        else:
            self._info_formatter = logging.Formatter(fmt="> %(message)s")
            self._warning_formatter = logging.Formatter(fmt=w_fmt, datefmt=datefmt)
            self._error_formatter = self._warning_formatter
            self._debug_formatter = self._warning_formatter

    def format(self, record: logging.LogRecord):
        if self.use_color:
            # Highlight all UUIDs
            if record.levelno > logging.WARNING:
                end_coloring = COLOR_ERROR
            elif record.levelno > logging.INFO:
                end_coloring = COLOR_WARNING
            elif record.levelno > logging.DEBUG:
                end_coloring = COLOR_INFO
            else:
                end_coloring = COLOR_DEBUG

            repl = f"{_byel}\\1{_r}{end_coloring}"
            try:
                record.msg = self.uuid_re.sub(repl, record.msg)
                if isinstance(record.args, dict):
                    for k, v in record.args.items():
                        record.args[k] = self.uuid_re.sub(repl, str(v))
                elif record.args:
                    args = tuple(self.uuid_re.sub(repl, str(a)) for a in record.args)
                    record.args = args
            except Exception as exc:
                # Basically, inform, but ignore
                print(f"Unable to colorize log message: {exc}")

        if record.levelno > logging.WARNING:
            return self._error_formatter.format(record)
        elif record.levelno > logging.INFO:
            return self._warning_formatter.format(record)
        elif record.levelno > logging.DEBUG:
            return self._info_formatter.format(record)
        return self._debug_formatter.format(record)


def _get_file_dict_config(
    logfile: str, console_enabled: bool, debug: bool, no_color: bool
) -> dict:
    # ensure that the logdir exists
    logdir = os.path.dirname(logfile)
    os.makedirs(logdir, exist_ok=True)
    log_handlers = ["logfile"]
    if console_enabled:
        log_handlers.append("console")

    return {
        "version": 1,
        "formatters": {
            "streamfmt": {
                "()": "funcx_endpoint.logging_config.FuncxConsoleFormatter",
                "debug": debug,
                "no_color": no_color,
            },
            "filefmt": {
                "format": DEFAULT_FORMAT,
                "datefmt": "%Y-%m-%d %H:%M:%S",
            },
        },
        "handlers": {
            "console": {
                "class": "logging.StreamHandler",
                "level": "DEBUG",
                "formatter": "streamfmt",
            },
            "logfile": {
                "class": "logging.handlers.RotatingFileHandler",
                "level": "DEBUG",
                "filename": logfile,
                "formatter": "filefmt",
                "maxBytes": 100 * 1024 * 1024,
                "backupCount": 1,
            },
        },
        "loggers": {
            "funcx_endpoint": {
                "level": "DEBUG" if debug else "INFO",
                "handlers": log_handlers,
            },
            # configure for the funcx SDK as well
            "funcx": {
                "level": "DEBUG" if debug else "WARNING",
                "handlers": log_handlers,
            },
        },
    }


def _get_stream_dict_config(debug: bool, no_color: bool) -> dict:
    return {
        "version": 1,
        "formatters": {
            "streamfmt": {
                "()": "funcx_endpoint.logging_config.FuncxConsoleFormatter",
                "debug": debug,
                "no_color": no_color,
            },
        },
        "handlers": {
            "console": {
                "class": "logging.StreamHandler",
                "level": "DEBUG" if debug else "INFO",
                "formatter": "streamfmt",
            }
        },
        "loggers": {
            "funcx_endpoint": {
                "level": "DEBUG",
                "handlers": ["console"],
            },
            # configure for the funcx SDK as well
            "funcx": {
                "level": "DEBUG" if debug else "WARNING",
                "handlers": ["console"],
            },
        },
    }


def add_trace_level() -> None:
    """This adds a trace level to the logging system.

    See https://stackoverflow.com/questions/2183233
    """

    TRACE = 5
    logging.TRACE = TRACE  # type: ignore[attr-defined]

    def logForLevel(self, message, *args, **kwargs):
        if self.isEnabledFor(TRACE):
            self._log(TRACE, message, args, **kwargs)

    def logToRoot(message, *args, **kwargs):
        logging.log(TRACE, message, *args, **kwargs)

    logging.addLevelName(TRACE, "TRACE")
    logging.getLoggerClass().trace = logForLevel  # type: ignore[attr-defined]
    logging.trace = logToRoot  # type: ignore[attr-defined]


def setup_logging(
    *,
    logfile: t.Optional[str] = None,
    console_enabled: bool = True,
    debug: bool = False,
    no_color: bool = False,
) -> None:
    add_trace_level()

    if logfile is not None:
        config = _get_file_dict_config(logfile, console_enabled, debug, no_color)
    else:
        config = _get_stream_dict_config(debug, no_color)

    logging.config.dictConfig(config)



from __future__ import annotations

import logging

import pika
import pika.channel
import pika.exceptions
from pika.adapters.blocking_connection import BlockingChannel

from .base import RabbitPublisherStatus

logger = logging.getLogger(__name__)


class ResultQueuePublisher:
    """ResultPublisher publishes results to a topic EXCHANGE_NAME, with
    the {endpoint_id}.results as a routing key.
    """

    def __init__(
        self,
        *,
        queue_info: dict,
    ):
        """
        Parameters
        ----------
        endpoint_uuid: str
            Endpoint UUID string used to identify the endpoint
        queue_info: pika.connection.Parameters
            Pika connection parameters to connect to RabbitMQ
        """
        self.queue_info = queue_info
        self._channel: BlockingChannel | None = None
        self._connection: pika.BlockingConnection | None = None
        # start closed ("connected" after connect)
        self.status = RabbitPublisherStatus.closed

        publish_kw = dict(**self.queue_info["queue_publish_kwargs"])
        if "properties" in publish_kw:
            publish_kw["properties"] = pika.BasicProperties(**publish_kw["properties"])
        self._publish_kwargs = publish_kw

    def __enter__(self):
        if self.status != RabbitPublisherStatus.connected:
            self.connect()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def connect(self):
        pika_params = pika.URLParameters(self.queue_info["connection_url"])
        pika_params.heartbeat = 0  # result_q is blocking; no heartbeats warranted
        conn = pika.BlockingConnection(pika_params)
        channel = conn.channel()
        channel.exchange_declare(passive=True, exchange=self.queue_info["exchange"])
        channel.queue_declare(passive=True, queue=self.queue_info["queue"])
        channel.confirm_delivery()

        self._connection = conn
        self._channel = channel
        self.status = RabbitPublisherStatus.connected

    def publish(self, message: bytes) -> None:
        """Publish message to RabbitMQ with the routing key to identify the message
        source.
        The channel specifies confirm_delivery and with `mandatory=True` this call
        will *block* until a delivery confirmation is received.

        Raises: Exception from pika if the message could not be delivered
        """
        if self._channel is None:
            raise ValueError("cannot publish() without first calling connect()")

        try:
            self._channel.basic_publish(body=message, **self._publish_kwargs)
        except pika.exceptions.AMQPError as err:
            logger.error(
                f"Unable to deliver message to exchange.\n"
                f"  Error text: {err}\n"
                f"  kwargs: {self._publish_kwargs}\n"
                f"  message: {message!r}"
            )
            raise

    def close(self):
        if self._connection and self._connection.is_open:
            self._connection.close()
        self.status = RabbitPublisherStatus.closed



"""
This module provides helpers for constructing the internal dict objects which get passed
around and ultimately converted into funcx_common.messagepack.Result objects

IDEALLY this would be refactored to produce and return Result objects directly, which
could then be passed around internally instead of raw dicts
but we don't have time to do that (even though it would be better (a lot better))
"""

from __future__ import annotations

import sys
import traceback
import types
import typing as t

from funcx.errors import MaxResultSizeExceeded
from funcx_endpoint.exceptions import CouldNotExecuteUserTaskError

INTERNAL_ERROR_CLASSES: tuple[type[Exception], ...] = (
    CouldNotExecuteUserTaskError,
    MaxResultSizeExceeded,
)


def _typed_excinfo() -> tuple[type[Exception], Exception, types.TracebackType]:
    return t.cast(
        t.Tuple[t.Type[Exception], Exception, types.TracebackType],
        sys.exc_info(),
    )


def _inner_traceback(tb: types.TracebackType, levels: int = 2) -> types.TracebackType:
    while levels > 0:
        tb = tb.tb_next if tb.tb_next is not None else tb
        levels -= 1
    return tb


def get_error_string(*, tb_levels: int = 2) -> str:
    exc_info = _typed_excinfo()
    exc_type, exc, tb = exc_info
    if isinstance(exc, INTERNAL_ERROR_CLASSES):
        return repr(exc)
    return "".join(
        traceback.format_exception(
            exc_type, exc, _inner_traceback(tb, levels=tb_levels)
        )
    )


def get_result_error_details() -> tuple[str, str]:
    _, error, _ = _typed_excinfo()
    # code, user_message
    if isinstance(error, INTERNAL_ERROR_CLASSES):
        return (error.__class__.__name__, f"remote error: {error}")
    return (
        "RemoteExecutionError",
        "An error occurred during the execution of this task",
    )



from __future__ import annotations

from parsl.utils import RepresentationMixin

from funcx_endpoint.executors import HighThroughputExecutor

_DEFAULT_EXECUTORS = [HighThroughputExecutor()]


class Config(RepresentationMixin):
    """Specification of FuncX configuration options.

    Parameters
    ----------

    executors : list of Executors
        A list of executors which serve as the backend for function execution.
        As of 0.2.2, this list should contain only one executor.
        Default: [HighThroughtputExecutor()]

    environment: str
        Environment the endpoint should connect to. Sets funcx_service_address and
        results_ws_uri unless they are also specified. If not specified, the endpoint
        connects to production.
        Default: None

    funcx_service_address: str | None
        URL address string of the funcX service to which the Endpoint should connect.
        Default: None

    results_ws_uri: str | None
        URL address string of the funcX websocket service passed to the funcX client.
        Default: None

    warn_about_url_mismatch: Bool
        Log a warning if funcx_service_address and results_ws_uri appear to point to
        different environments.
        Default: False

    heartbeat_period: int (seconds)
        The interval at which heartbeat messages are sent from the endpoint to the
        funcx-web-service
        Default: 30s

    heartbeat_threshold: int (seconds)
        Seconds since the last hearbeat message from the funcx-web-service after which
        the connection is assumed to be disconnected.
        Default: 120s

    stdout : str
        Path where the endpoint's stdout should be written
        Default: ./interchange.stdout

    stderr : str
        Path where the endpoint's stderr should be written
        Default: ./interchange.stderr

    detach_endpoint : Bool
        Should the endpoint deamon be run as a detached process? This is good for
        a real edge node, but an anti-pattern for kubernetes pods
        Default: True

    log_dir : str
        Optional path string to the top-level directory where logs should be written to.
        Default: None
    """

    def __init__(
        self,
        # Execution backed
        executors: list = _DEFAULT_EXECUTORS,
        # Connection info
        environment: str | None = None,
        funcx_service_address=None,
        results_ws_uri=None,
        warn_about_url_mismatch=False,
        # Tuning info
        heartbeat_period=30,
        heartbeat_threshold=120,
        detach_endpoint=True,
        # Logging info
        log_dir=None,
        stdout="./endpoint.log",
        stderr="./endpoint.log",
    ):

        # Execution backends
        self.executors = executors  # List of executors

        # Connection info
        self.environment = environment
        self.funcx_service_address = funcx_service_address
        self.results_ws_uri = results_ws_uri
        self.warn_about_url_mismatch = warn_about_url_mismatch

        # Tuning info
        self.heartbeat_period = heartbeat_period
        self.heartbeat_threshold = heartbeat_threshold
        self.detach_endpoint = detach_endpoint

        # Logging info
        self.log_dir = log_dir
        self.stdout = stdout
        self.stderr = stderr



import logging
import math
import random

log = logging.getLogger(__name__)


def naive_scheduler(
    task_qs, outstanding_task_count, max_workers, old_worker_map, to_die_list
):
    """
    Return two items (as one tuple)
        dict kill_list :: KILL [(worker_type, num_kill), ...]
        dict create_list :: CREATE [(worker_type, num_create), ...]

    In this scheduler model, there is minimum 1 instance of each nonempty task queue.
    """

    log.trace("Entering scheduler...")
    log.trace(f"old_worker_map: {old_worker_map}")
    q_sizes = {}
    q_types = []
    new_worker_map = {}

    # Sum the size of each *available* (unblocked) task queue
    sum_q_size = 0
    for q_type in outstanding_task_count:
        q_types.append(q_type)
        q_size = outstanding_task_count[q_type]
        sum_q_size += q_size
        q_sizes[q_type] = q_size

    if sum_q_size > 0:
        log.info(f"Total number of tasks is {sum_q_size}")

        # Set proportions of workers equal to the proportion of queue size.
        for q_type in q_sizes:
            ratio = q_sizes[q_type] / sum_q_size
            new_worker_map[q_type] = min(
                int(math.floor(ratio * max_workers)), q_sizes[q_type]
            )

        # CLEANUP: Assign the difference here to any random worker. Should be small.
        # log.debug("Temporary new worker map: {}".format(new_worker_map))

        # Check the difference
        tmp_sum_q_size = sum(new_worker_map.values())
        difference = 0
        if sum_q_size > tmp_sum_q_size:
            difference = min(max_workers - tmp_sum_q_size, sum_q_size - tmp_sum_q_size)
        log.debug(f"Offset difference: {difference}")
        log.debug(f"Queue Types: {q_types}")

        if len(q_types) > 0:
            while difference > 0:
                win_q = random.choice(q_types)
                if q_sizes[win_q] > new_worker_map[win_q]:
                    new_worker_map[win_q] += 1
                    difference -= 1

    return new_worker_map



from __future__ import annotations

import logging
import multiprocessing
import os
import platform
import queue
import signal
import sys
import threading
import time
import typing as t

# multiprocessing.Event is a method, not a class
# to annotate, we need the "real" class
# see: https://github.com/python/typeshed/issues/4266
from multiprocessing.synchronize import Event as EventType

import dill
from funcx_common.messagepack import pack
from funcx_common.messagepack.message_types import Result, ResultErrorDetails
from parsl.version import VERSION as PARSL_VERSION
from retry.api import retry_call

import funcx_endpoint.endpoint.utils.config
from funcx import __version__ as funcx_sdk_version
from funcx.sdk.client import FuncXClient
from funcx_endpoint import __version__ as funcx_endpoint_version
from funcx_endpoint.endpoint.messages_compat import (
    try_convert_from_messagepack,
    try_convert_to_messagepack,
)
from funcx_endpoint.endpoint.rabbit_mq import ResultQueuePublisher, TaskQueueSubscriber
from funcx_endpoint.endpoint.register_endpoint import register_endpoint
from funcx_endpoint.endpoint.result_store import ResultStore
from funcx_endpoint.executors.high_throughput.mac_safe_queue import mpQueue

log = logging.getLogger(__name__)

LOOP_SLOWDOWN = 0.0  # in seconds
HEARTBEAT_CODE = (2**32) - 1
PKL_HEARTBEAT_CODE = dill.dumps(HEARTBEAT_CODE)


class EndpointInterchange:
    """Interchange is a task orchestrator for distributed systems.

    1. Asynchronously queue large volume of tasks (>100K)
    2. Allow for workers to join and leave the union
    3. Detect workers that have failed using heartbeats
    4. Service single and batch requests from workers
    5. Be aware of requests worker resource capacity,
       eg. schedule only jobs that fit into walltime.

    """

    def __init__(
        self,
        config: funcx_endpoint.endpoint.utils.config.Config,
        reg_info: t.Tuple[t.Dict, t.Dict] = None,
        logdir=".",
        endpoint_id=None,
        endpoint_dir=".",
        endpoint_name="default",
        funcx_client: FuncXClient | None = None,
        result_store: ResultStore | None = None,
    ):
        """
        Parameters
        ----------
        config : funcx.Config object
             Funcx config object that describes how compute should be provisioned

        reg_info : tuple[Dict, Dict]
             Tuple of connection info for task_queue and result_queue
             Connection parameters to connect to the service side RabbitMQ pipes
             Optional: If not supplied, the endpoint will use a retry loop to
             attempt registration periodically.

        logdir : str
             Parsl log directory paths. Logs and temp files go here. Default: '.'

        endpoint_id : str
             Identity string that identifies the endpoint to the broker

        endpoint_dir : str
             Endpoint directory path to store registration info in

        endpoint_name : str
             Name of endpoint

        funcx_client_options : Dict
             FuncXClient initialization options
        """
        self.logdir = logdir
        log.info(
            "Initializing EndpointInterchange process with Endpoint ID: {}".format(
                endpoint_id
            )
        )
        self.config = config

        self.endpoint_dir = endpoint_dir
        self.endpoint_name = endpoint_name

        if funcx_client is None:
            funcx_client = FuncXClient()
        self.funcx_client = funcx_client

        self.initial_registration_complete = False
        self.task_q_info, self.result_q_info = None, None
        if reg_info:
            self.task_q_info, self.result_q_info = reg_info
            self.initial_registration_complete = True

        self.time_to_quit = False
        self.heartbeat_period = self.config.heartbeat_period

        self.pending_task_queue: multiprocessing.Queue = multiprocessing.Queue()

        self._quiesce_event = multiprocessing.Event()
        self._kill_event = multiprocessing.Event()

        if result_store is None:
            result_store = ResultStore(endpoint_dir=endpoint_dir)
        self.result_store = result_store

        self.endpoint_id = endpoint_id

        self.current_platform = {
            "parsl_v": PARSL_VERSION,
            "python_v": "{}.{}.{}".format(
                sys.version_info.major, sys.version_info.minor, sys.version_info.micro
            ),
            "os": platform.system(),
            "hname": platform.node(),
            "funcx_sdk_version": funcx_sdk_version,
            "funcx_endpoint_version": funcx_endpoint_version,
            "registration": self.endpoint_id,
            "dir": os.getcwd(),
        }

        log.info(f"Platform info: {self.current_platform}")
        try:
            self.load_config()
        except Exception:
            log.exception("Caught exception")
            raise

        self._test_start = False

    def load_config(self):
        """Load the config"""
        log.info("Loading endpoint local config")

        self.results_passthrough = mpQueue()
        self.executors: dict[str, funcx_endpoint.executors.HighThroughputExecutor] = {}
        for executor in self.config.executors:
            log.info(f"Initializing executor: {executor.label}")
            executor.funcx_service_address = self.config.funcx_service_address
            if not executor.endpoint_id:
                executor.endpoint_id = self.endpoint_id
            else:
                if not executor.endpoint_id == self.endpoint_id:
                    eep_id = f"Executor({executor.endpoint_id})"
                    sep_id = f"Interchange({self.endpoint_id})"
                    raise Exception(f"InconsistentEndpointId: {eep_id} != {sep_id}")
            self.executors[executor.label] = executor
            if executor.run_dir is None:
                executor.run_dir = self.logdir

    def start_executors(self):
        log.info("Starting Executors")
        for executor in self.config.executors:
            if hasattr(executor, "passthrough") and executor.passthrough is True:
                executor.start(results_passthrough=self.results_passthrough)

    def register_endpoint(self):
        reg_info = register_endpoint(
            self.funcx_client, self.endpoint_id, self.endpoint_dir, self.endpoint_name
        )
        self.task_q_info, self.result_q_info = reg_info

    def migrate_tasks_to_internal(
        self,
        connection_params: dict,
        endpoint_uuid: str,
        pending_task_queue: multiprocessing.Queue,
        quiesce_event: EventType,
    ) -> multiprocessing.Process:
        """Pull tasks from the incoming tasks 0mq pipe onto the internal
        pending task queue

        Parameters:
        -----------
        connection_params: pika.connection.Parameters
              Connection params to connect to the service side Tasks queue

        endpoint_uuid: endpoint_uuid str

        pending_task_queue: multiprocessing.Queue
              Internal queue to which tasks should be migrated

        quiesce_event : EventType
              Event to let the thread know when it is time to die.
        """
        try:
            log.info(f"Starting the TaskQueueSubscriber as {endpoint_uuid}")
            task_q_proc = TaskQueueSubscriber(
                queue_info=connection_params,
                external_queue=pending_task_queue,
                quiesce_event=quiesce_event,
                endpoint_id=endpoint_uuid,
            )
            task_q_proc.start()
        except Exception:
            log.exception("Unhandled exception in TaskQueueSubscriber")
            raise

        return task_q_proc

    def quiesce(self):
        """Temporarily stop everything on the interchange in order to reach a consistent
        state before attempting to start again. This must be called on the main thread
        """
        log.info("Interchange Quiesce in progress (stopping and joining processes)")
        self._quiesce_event.set()

        log.info("Waiting for quiesce complete")
        self._task_puller_proc.join()

        log.info("Quiesce done")

    def stop(self):
        """Prepare the interchange for shutdown"""
        log.info("Shutting down EndpointInterchange")

        # kill_event must be set before quiesce_event because we need to guarantee that
        # once the quiesce is complete, the interchange will not try to start again
        self._kill_event.set()
        self._quiesce_event.set()

    def cleanup(self):
        for label in self.executors:
            self.executors[label].shutdown()

    def handle_sigterm(self, sig_num, curr_stack_frame):
        log.warning("Received SIGTERM, setting termination flag.")
        self.time_to_quit = True

    def start(self):
        """Start the Interchange"""
        log.info("Starting EndpointInterchange")

        signal.signal(signal.SIGTERM, self.handle_sigterm)
        signal.signal(signal.SIGQUIT, self.handle_sigterm)  # hint: Ctrl+\
        # Intentionally ignoring SIGINT for now, as we're unstable enough to
        # warrant Python's default developer-friendly Ctrl+C handling

        self._quiesce_event.clear()
        self._kill_event.clear()

        # NOTE: currently we only start the executors once because
        # the current behavior is to keep them running decoupled while
        # the endpoint is waiting for reconnection
        self.start_executors()

        while not self._kill_event.is_set():
            if self._quiesce_event.is_set():
                log.warning("Interchange will retry connecting in 5s")
                time.sleep(5)
                self._quiesce_event.clear()
            else:
                log.debug("Starting threads and main loop")

            try:
                self._start_threads_and_main()
            except Exception:
                log.exception("Unhandled exception in main kernel.")

            self.quiesce()

        self.cleanup()
        log.info("EndpointInterchange shutdown complete.")

    def _start_threads_and_main(self):
        # re-register on every loop start
        if not self.initial_registration_complete:
            # Register the endpoint
            log.info("Running endpoint registration retry loop")
            retry_call(self.register_endpoint, delay=10, max_delay=300, backoff=1.2)
            log.info("Endpoint registered with UUID: (FIXME FIXME FIXME)")

        self.initial_registration_complete = False

        self._task_puller_proc = self.migrate_tasks_to_internal(
            self.task_q_info,
            self.endpoint_id,
            self.pending_task_queue,
            self._quiesce_event,
        )

        self._main_loop()

    def _main_loop(self):
        """
        This is the "kernel" of the endpoint interchange process.  Conceptually, there
        are three actions of consequence: forward task messages to the executors,
        forward results from the executors back to the funcx web services (RMQ), and
        forward any previous results that may have failed to send previously (e.g., if
        a RMQ connection was dropped).

        We accomplish this via three threads, one each for each task.

        Of special note is that this kernel does not try very hard to handle the
        non-happy path.  If an error occurs that is too "unhappy" (e.g., communication
        with RMQ fails), the catch-all solution is to "reboot" and try again.  That is,
        this method exits, communication is shutdown, started again, and we restart
        this method.
        """
        log.debug("_main_loop begins")

        results_publisher = ResultQueuePublisher(
            queue_info=self.result_q_info,
        )

        with results_publisher:
            executor = list(self.executors.values())[0]

            num_tasks_forwarded = 0
            num_results_forwarded = 0

            def process_stored_results():
                # Handle any previously stored results, either from a previous run or
                # from a quarter of a second-ago.  Basically, check every second
                # (.wait()) if there are any items on disk to be sent.  If there are,
                # don't treat them any differently than "fresh" results: put the into
                # the same multiprocessing queue as results incoming directly from
                # the executors.  The normal processing by `process_pending_results()`
                # will take over from there.
                while not self._quiesce_event.wait(timeout=1):
                    for task_id, packed_result in self.result_store:
                        if self._quiesce_event.is_set():
                            # important to check every iteration as well, so as not to
                            # potentially hang up the shutdown procedure
                            return
                        log.debug("Retrieved stored result (%s)", task_id)
                        msg = {"task_id": task_id, "message": packed_result}
                        self.result_store.discard(task_id)
                        self.results_passthrough.put(msg)
                log.debug("Exit process-stored-results thread.")

                log.debug("Exit process-stored-results thread.")

            def process_pending_tasks():
                # Pull tasks from upstream (RMQ) and send them down the ZMQ pipe to the
                # funcx-manager.  In terms of shutting down (or "rebooting") gracefully,
                # iterate once a second whether or not a task has arrived.
                nonlocal num_tasks_forwarded
                while not self._quiesce_event.is_set():
                    if self.time_to_quit:
                        self.stop()
                        continue  # nominally == break; but let event do it

                    try:
                        incoming_task = self.pending_task_queue.get(timeout=1)
                        task = try_convert_from_messagepack(incoming_task)
                        executor.submit_raw(task)
                        num_tasks_forwarded += 1  # Safe given GIL

                    except queue.Empty:
                        continue

                    except Exception:
                        log.exception("Unhandled issue while waiting for pending tasks")

                log.debug("Exit process-pending-tasks thread.")

            def process_pending_results():
                # Forward incoming results from the funcx-manager to the funcx-services.
                # For graceful handling of shutdown (or "reboot"), wait up to a second
                # for incoming results before iterating the loop regardless.
                nonlocal num_results_forwarded
                while not self._quiesce_event.is_set():
                    try:
                        result = self.results_passthrough.get(timeout=1)
                        task_id = result["task_id"]
                        packed_result = result["message"]

                    except queue.Empty:
                        # Empty queue!  Let's see if we have any prior results to send
                        continue

                    except Exception as exc:
                        log.debug(f"Invalid message received: {result}")
                        log.warning(
                            f"Invalid message received: no task_id.  Ignoring. {exc}"
                        )
                        continue

                    try:
                        # This either works or it doesn't; if it doesn't to serialize
                        # the to an execption and send _that_
                        # will be a packed EPStatusReport or Result
                        message = try_convert_to_messagepack(packed_result)

                    except Exception as exc:
                        log.exception(
                            f"Unable to parse result message for task {task_id}."
                            "   Marking task as failed."
                        )

                        kwargs = {
                            "task_id": task_id,
                            "data": packed_result,
                            "error_details": ResultErrorDetails(
                                code=0,
                                user_message=(
                                    "Endpoint failed to serialize."
                                    f"  Exception text: {exc}"
                                ),
                            ),
                        }
                        message = pack(Result(**kwargs))

                    if task_id:
                        log.debug(f"Forwarding result for task {task_id}")

                    try:
                        results_publisher.publish(message)
                        num_results_forwarded += 1  # Safe given GIL

                    except Exception:
                        # Publishing didn't work -- quiesce and see if a simple restart
                        # fixes the issue.
                        self._quiesce_event.set()

                        log.exception("Something broke while forwarding results")
                        if task_id:
                            log.info("Storing result for later: %s", task_id)
                            self.result_store[task_id] = packed_result
                        continue  # just be explicit

                log.debug("Exit process-pending-results thread.")

            stored_processor_thread = threading.Thread(
                target=process_stored_results, name="Stored Result Handler"
            )
            task_processor_thread = threading.Thread(
                target=process_pending_tasks, name="Pending Task Handler"
            )
            result_processor_thread = threading.Thread(
                target=process_pending_results, name="Pending Result Handler"
            )
            stored_processor_thread.start()
            task_processor_thread.start()
            result_processor_thread.start()

            log.debug("_main_loop entered running state")
            last_t, last_r = 0, 0
            while not self._quiesce_event.wait(self.heartbeat_period):
                # Possibly TOCTOU here, but we don't need to be super precise.  The
                # point here is to mention "still alive" and that we're still working
                num_t, num_r = num_tasks_forwarded, num_results_forwarded
                log.debug(
                    "Heartbeat.  Approximate Tasks and Results forwarded since last "
                    "heartbeat: %s (T), %s (R)",
                    num_t - last_t,
                    num_r - last_r,
                )
                last_t, last_r = num_t, num_r

            # The timeouts aren't nominally necessary because if the above loop has
            # quit, then the _quiesce_event is set, and both threads check that event
            # every internal iteration.  But "for kicks."
            stored_processor_thread.join(timeout=5)
            task_processor_thread.join(timeout=5)
            result_processor_thread.join(timeout=5)



from __future__ import annotations

import os
import pathlib
import sys
import typing as t


class ResultStore:
    """
    Persists bytes to disk by key via a dict()-like API.

    The ResultStore is a generic byte-storage data structure that stores and
    retrieves bytes to the filesystem.  Each payload is stored on the
    filesystem as a single file, named by the associated key.  Payloads are
    stored in a subdirectory of the init path, named "unacked_results".

    A typical interaction might look like:

        >>> rs = ResultStore(endpoint_dir="some_dir")
        >>> rs["some_key_01"] = b"some_payload_01"

    Access a single result by the result key

        >>> stored_bytes = rs["some_key_01"]  # raises if key does not exist
        >>> stored_bytes = rs.get("some_key_01")  # returns None if key doesn't exist

    Iterate all currently stored items via iteration:

        >>> for key_str, stored_bytes in rs:

    Discard stored items via `.pop()`, `.remove()`, or `.discard()`

        >>> rs.pop("some_key_01")  # also returns result, but ignored here
        >>> rs.pop("some_key_01")  # will raise as key does not exist
        >>> rs.discard("some_key_02")  # will not raise if key does not exist

    Completely empty the ResultStore with `.clear()`.  All non-hidden files
    (those not beginning with a dot) in the unacked_results directory will be
    removed from the filesystem.

        >>> rs.clear()
    """

    def __init__(self, endpoint_dir: str | pathlib.Path):
        self.endpoint_dir = pathlib.Path(endpoint_dir)
        self.data_path = self.endpoint_dir / "unacked_results"
        self.data_path.mkdir(exist_ok=True)

        self.data_path.chmod(mode=0o0700)
        test_path = self.data_path / ".test-reading-and-writing.txt"
        test_path.write_bytes(b"Verify write-ability")
        if not any(i for i in self.data_path.glob(".test*")):
            msg = (
                f"Unable to list files; does {self.data_path} have the correct"
                " permissions?"
            )
            raise PermissionError(msg)
        test_path.unlink()

    def __contains__(self, key: str) -> bool:
        normpath = os.path.normpath(key)
        return (self.data_path / normpath).exists()

    def __iter__(self) -> t.Iterable[tuple[str, bytes]]:
        for rp in self._iter_result_paths():
            yield rp.name, rp.read_bytes()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(data_dir: {self.data_path})"

    def __getitem__(self, key):
        rp = self.data_path / key
        return rp.read_bytes()

    def __setitem__(self, key: str, payload: bytes) -> None:
        """
        Store bytes to disk by filename 'key'
        """
        result_path = self.data_path / os.path.normpath(key)
        result_path.write_bytes(payload)

    def __delitem__(self, key: str) -> None:
        """
        Remove the payload corresponding to key from the store.

        If there is no payload for key, raise a FileNotFoundError.
        """
        (self.data_path / key).unlink()

    def discard(self, key: str) -> None:
        """
        Discard the requested task result.  Will not raise if backing file does
        not exist.

        Parameters
        ----------
        key - the key for the result
        """
        if sys.version_info < (3, 8):
            try:
                (self.data_path / key).unlink()
            except FileNotFoundError:
                pass
        else:
            (self.data_path / key).unlink(missing_ok=True)

    def _iter_result_paths(self) -> t.Iterable[pathlib.Path]:
        yield from self.data_path.glob("[!.]*")

    def get(self, key: str, default=None) -> bytes:
        """
        Retrieve result from the data directory.

        Raises
        ------
        FileNotFoundError when backing file does not exist

        Returns
        -------
        message
        """
        try:
            rp = self.data_path / key
            return rp.read_bytes()
        except FileNotFoundError:
            return default

    def pop(self, key: str, *args, **kwargs) -> bytes:
        """
        Retrieve and remove a result from the data directory.

        Raises
        ------
        FileNotFoundError when backing file does not exist

        Returns
        -------
        message
        """
        try:
            payload = self[key]
            self.discard(key)
            return payload
        except FileNotFoundError:
            if args:
                return args[0]
            elif "default" in kwargs:
                return kwargs["default"]
            raise

    def clear(self) -> None:
        """
        Remove all stored items from the storage directory.  This iterates the
        storage for all non-hidden files (those that don't begin with a dot)
        and safely unlinks them.
        """
        if sys.version_info < (3, 8):
            for result_path in self._iter_result_paths():
                try:
                    result_path.unlink()
                except FileNotFoundError:
                    pass
        else:
            for result_path in self._iter_result_paths():
                result_path.unlink(missing_ok=True)



import collections
import logging
import queue
import random

log = logging.getLogger(__name__)
log.info("Interchange task dispatch started")


def naive_interchange_task_dispatch(
    interesting_managers,
    pending_task_queue,
    ready_manager_queue,
    scheduler_mode="hard",
    cold_routing=False,
):
    """
    This is an initial task dispatching algorithm for interchange.
    It returns a dictionary, whose key is manager, and the value is the list of tasks
    to be sent to manager, and the total number of dispatched tasks.
    """
    if scheduler_mode == "hard":
        return dispatch(
            interesting_managers,
            pending_task_queue,
            ready_manager_queue,
            scheduler_mode="hard",
        )

    elif scheduler_mode == "soft":
        task_dispatch, dispatched_tasks = {}, 0
        loops = ["warm"] if not cold_routing else ["warm", "cold"]
        for loop in loops:
            task_dispatch, dispatched_tasks = dispatch(
                interesting_managers,
                pending_task_queue,
                ready_manager_queue,
                scheduler_mode="soft",
                loop=loop,
                task_dispatch=task_dispatch,
                dispatched_tasks=dispatched_tasks,
            )
        return task_dispatch, dispatched_tasks


def dispatch(
    interesting_managers,
    pending_task_queue,
    ready_manager_queue,
    scheduler_mode="hard",
    loop="warm",
    task_dispatch=None,
    dispatched_tasks=0,
):
    """
    This is the core task dispatching algorithm for interchange.
    The algorithm depends on the scheduler mode and which loop.
    """
    if not task_dispatch:
        task_dispatch = {}
    if interesting_managers:
        shuffled_managers = list(interesting_managers)
        random.shuffle(shuffled_managers)
        for manager in shuffled_managers:
            tasks_inflight = ready_manager_queue[manager]["total_tasks"]
            real_capacity = min(
                ready_manager_queue[manager]["free_capacity"]["total_workers"],
                ready_manager_queue[manager]["max_worker_count"] - tasks_inflight,
            )
            if real_capacity and ready_manager_queue[manager]["active"]:
                if scheduler_mode == "hard":
                    tasks, tids = get_tasks_hard(
                        pending_task_queue, ready_manager_queue[manager], real_capacity
                    )
                else:
                    tasks, tids = get_tasks_soft(
                        pending_task_queue,
                        ready_manager_queue[manager],
                        real_capacity,
                        loop=loop,
                    )
                if tasks:
                    log.debug(f"Got {len(tasks)} tasks from queue")
                    for task_type in tids:
                        # This line is a set update, not dict update
                        ready_manager_queue[manager]["tasks"][task_type].update(
                            tids[task_type]
                        )
                    log.debug(
                        "The tasks on manager {} is {}".format(
                            manager, ready_manager_queue[manager]["tasks"]
                        )
                    )
                    ready_manager_queue[manager]["total_tasks"] += len(tasks)
                    if manager not in task_dispatch:
                        task_dispatch[manager] = []
                    task_dispatch[manager] += tasks
                    dispatched_tasks += len(tasks)
                    log.debug(f"Assigned tasks {tids} to manager {manager}")
                if ready_manager_queue[manager]["free_capacity"]["total_workers"] > 0:
                    log.trace(
                        "Manager {} still has free_capacity {}".format(
                            manager,
                            ready_manager_queue[manager]["free_capacity"][
                                "total_workers"
                            ],
                        )
                    )
                else:
                    log.debug(f"Manager {manager} is now saturated")
                    interesting_managers.remove(manager)
            else:
                interesting_managers.remove(manager)

    log.trace(
        "The task dispatch of {} loop is {}, in total {} tasks".format(
            loop, task_dispatch, dispatched_tasks
        )
    )
    return task_dispatch, dispatched_tasks


def get_tasks_hard(pending_task_queue, manager_ads, real_capacity):
    tasks = []
    tids = collections.defaultdict(set)
    task_type = manager_ads["worker_type"]
    if not task_type:
        log.warning(
            "Using hard scheduler mode but with manager worker type unset. "
            "Use soft scheduler mode. Set this in the config."
        )
        return tasks, tids
    if task_type not in pending_task_queue:
        log.trace(f"No task of type {task_type}. Exiting task fetching.")
        return tasks, tids

    # dispatch tasks of available types on manager
    if task_type in manager_ads["free_capacity"]["free"]:
        while manager_ads["free_capacity"]["free"][task_type] > 0 and real_capacity > 0:
            try:
                x = pending_task_queue[task_type].get(block=False)
            except queue.Empty:
                break
            else:
                log.debug(f"Get task {x}")
                tasks.append(x)
                tids[task_type].add(x["task_id"])
                manager_ads["free_capacity"]["free"][task_type] -= 1
                manager_ads["free_capacity"]["total_workers"] -= 1
                real_capacity -= 1

    # dispatch tasks to unused slots based on the manager type
    log.trace("Second round of task fetching in hard mode")
    while manager_ads["free_capacity"]["free"]["unused"] > 0 and real_capacity > 0:
        try:
            x = pending_task_queue[task_type].get(block=False)
        except queue.Empty:
            break
        else:
            log.debug(f"Get task {x}")
            tasks.append(x)
            tids[task_type].add(x["task_id"])
            manager_ads["free_capacity"]["free"]["unused"] -= 1
            manager_ads["free_capacity"]["total_workers"] -= 1
            real_capacity -= 1
    return tasks, tids


def get_tasks_soft(pending_task_queue, manager_ads, real_capacity, loop="warm"):
    tasks = []
    tids = collections.defaultdict(set)

    # Warm routing to dispatch tasks
    if loop == "warm":
        for task_type in manager_ads["free_capacity"]["free"]:
            # Dispatch tasks that are of the available container types on the manager
            if task_type != "unused":
                type_inflight = len(manager_ads["tasks"].get(task_type, set()))
                type_capacity = min(
                    manager_ads["free_capacity"]["free"][task_type],
                    manager_ads["free_capacity"]["total"][task_type] - type_inflight,
                )
                while (
                    manager_ads["free_capacity"]["free"][task_type] > 0
                    and real_capacity > 0
                    and type_capacity > 0
                ):
                    try:
                        if task_type not in pending_task_queue:
                            break
                        x = pending_task_queue[task_type].get(block=False)
                    except queue.Empty:
                        break
                    else:
                        log.debug(f"Get task {x}")
                        tasks.append(x)
                        tids[task_type].add(x["task_id"])
                        manager_ads["free_capacity"]["free"][task_type] -= 1
                        manager_ads["free_capacity"]["total_workers"] -= 1
                        real_capacity -= 1
                        type_capacity -= 1
            # Dispatch tasks to unused container slots on the manager
            else:
                task_types = list(pending_task_queue.keys())
                random.shuffle(task_types)
                for task_type in task_types:
                    while (
                        manager_ads["free_capacity"]["free"]["unused"] > 0
                        and manager_ads["free_capacity"]["total_workers"] > 0
                        and real_capacity > 0
                    ):
                        try:
                            x = pending_task_queue[task_type].get(block=False)
                        except queue.Empty:
                            break
                        else:
                            log.debug(f"Get task {x}")
                            tasks.append(x)
                            tids[task_type].add(x["task_id"])
                            manager_ads["free_capacity"]["free"]["unused"] -= 1
                            manager_ads["free_capacity"]["total_workers"] -= 1
                            real_capacity -= 1
        return tasks, tids

    # Cold routing round: allocate tasks of random types
    # to workers that are of different types on the manager
    # This will possibly cause container switching on the manager
    # This is needed to avoid workers being idle for too long
    # Potential issues may be that it could kill containers of short tasks frequently
    # Tune cold_routing_interval in the config to balance such a tradeoff
    log.debug("Cold function routing!")
    task_types = list(pending_task_queue.keys())
    random.shuffle(task_types)
    for task_type in task_types:
        while manager_ads["free_capacity"]["total_workers"] > 0 and real_capacity > 0:
            try:
                x = pending_task_queue[task_type].get(block=False)
            except queue.Empty:
                break
            else:
                log.debug(f"Get task {x}")
                tasks.append(x)
                tids[task_type].add(x["task_id"])
                manager_ads["free_capacity"]["total_workers"] -= 1
                real_capacity -= 1
    return tasks, tids



# single source of truth for package version,
# see https://packaging.python.org/en/latest/single_source_version/
__version__ = "1.0.4"

# TODO: remove after a `funcx` release
# this variable is needed because it's imported by `funcx` to do the version check
VERSION = __version__

# app name to send as part of requests
app_name = f"funcX Endpoint v{__version__}"



import glob
import json
import logging
import os
import pathlib
import shutil
import signal
import sys
import typing
import uuid
from string import Template

import click
import daemon
import daemon.pidfile
import psutil
import texttable
from globus_sdk import GlobusAPIError, NetworkError

from funcx.errors import FuncxResponseError
from funcx.sdk.client import FuncXClient
from funcx_endpoint.endpoint import default_config as endpoint_default_config
from funcx_endpoint.endpoint.interchange import EndpointInterchange
from funcx_endpoint.endpoint.register_endpoint import register_endpoint
from funcx_endpoint.endpoint.result_store import ResultStore
from funcx_endpoint.logging_config import setup_logging

log = logging.getLogger(__name__)


_DEFAULT_FUNCX_DIR = str(pathlib.Path.home() / ".funcx")


class Endpoint:
    """
    Endpoint is primarily responsible for configuring, launching and stopping
    the Endpoint.
    """

    def __init__(self, funcx_dir=None, debug=False):
        """Initialize the Endpoint

        Parameters
        ----------

        funcx_dir: str
            Directory path to the root of the funcx dirs. Usually ~/.funcx.

        debug: Bool
            Enable debug logging. Default: False
        """
        self.debug = debug
        self.funcx_dir = funcx_dir if funcx_dir is not None else _DEFAULT_FUNCX_DIR
        self.name = "default"

    @property
    def _endpoint_dir(self):
        return os.path.join(self.funcx_dir, self.name)

    @property
    def _config_file_path(self):
        return os.path.join(self._endpoint_dir, "config.py")

    def init_endpoint_dir(self, endpoint_config=None):
        """Initialize a clean endpoint dir
        Returns if an endpoint_dir already exists

        Parameters
        ----------
        endpoint_config : str
            Path to a config file to be used instead of the funcX default config file
        """

        log.debug(f"Creating endpoint dir {self._endpoint_dir}")
        os.makedirs(self._endpoint_dir, exist_ok=True)

        endpoint_config_target_file = self._config_file_path
        if endpoint_config:
            shutil.copyfile(endpoint_config, endpoint_config_target_file)
            return self._endpoint_dir

        endpoint_config = endpoint_default_config.__file__
        with open(endpoint_config) as r:
            endpoint_config_template = Template(r.read())

        endpoint_config_template = endpoint_config_template.substitute(name=self.name)
        with open(endpoint_config_target_file, "w") as w:
            w.write(endpoint_config_template)

        return self._endpoint_dir

    def configure_endpoint(self, name, endpoint_config):
        self.name = name

        if not os.path.exists(self._endpoint_dir):
            self.init_endpoint_dir(endpoint_config=endpoint_config)
            print(
                f"A default profile has been create for <{self.name}> "
                f"at {self._config_file_path}"
            )
            print("Configure this file and try restarting with:")
            print(f"    $ funcx-endpoint start {self.name}")
        else:
            print(f"config dir <{self.name}> already exsits")
            raise Exception("ConfigExists")

    def init_endpoint(self):
        """Setup funcx dirs and default endpoint config files

        TO-DO : Every mechanism that will update the config file, must be using a
        locking mechanism, ideally something like fcntl [1]
        to ensure that multiple endpoint invocations do not mangle the funcx config
        files or the lockfile module.

        [1] https://docs.python.org/3/library/fcntl.html
        """
        # construct client to force login flow
        # FIXME: `funcx` should provide a cleaner way of doing login
        FuncXClient(do_version_check=False)

        if os.path.exists(self.funcx_dir):
            click.confirm(
                "Are you sure you want to initialize this directory? "
                f"This will erase everything in {self.funcx_dir}",
                abort=True,
            )
            log.info(f"Wiping all current configs in {self.funcx_dir}")
            backup_dir = self.funcx_dir + ".bak"
            try:
                log.debug(f"Removing old backups in {backup_dir}")
                shutil.rmtree(backup_dir)
            except OSError:
                pass
            os.renames(self.funcx_dir, backup_dir)

        try:
            os.makedirs(self.funcx_dir, exist_ok=True)
        except Exception as e:
            print(f"[FuncX] Caught exception during registration {e}")

    @staticmethod
    def check_endpoint_json(endpoint_json, endpoint_uuid):
        if os.path.exists(endpoint_json):
            with open(endpoint_json) as fp:
                log.debug("Connection info loaded from prior registration record")
                reg_info = json.load(fp)
                endpoint_uuid = reg_info["endpoint_id"]
        elif not endpoint_uuid:
            endpoint_uuid = str(uuid.uuid4())
        return endpoint_uuid

    def start_endpoint(
        self,
        name,
        endpoint_uuid,
        endpoint_config,
        log_to_console: bool,
        no_color: bool,
    ):
        self.name = name

        endpoint_dir = os.path.join(self.funcx_dir, self.name)
        endpoint_json = os.path.join(endpoint_dir, "endpoint.json")

        # This is to ensure that at least 1 executor is defined
        if not endpoint_config.config.executors:
            raise Exception(
                f"Endpoint config file at {endpoint_dir} is missing "
                "executor definitions"
            )

        funcx_client_options = {
            "funcx_service_address": endpoint_config.config.funcx_service_address,
            "results_ws_uri": endpoint_config.config.results_ws_uri,
            "environment": endpoint_config.config.environment,
            "warn_about_url_mismatch": endpoint_config.config.warn_about_url_mismatch,
        }
        funcx_client = FuncXClient(**funcx_client_options)

        endpoint_uuid = self.check_endpoint_json(endpoint_json, endpoint_uuid)

        log.info(f"Starting endpoint with uuid: {endpoint_uuid}")

        pid_file = os.path.join(endpoint_dir, "daemon.pid")
        pid_check = self.check_pidfile(pid_file)
        # if the pidfile exists, we should return early because we don't
        # want to attempt to create a new daemon when one is already
        # potentially running with the existing pidfile
        if pid_check["exists"]:
            if pid_check["active"]:
                log.info("Endpoint is already active")
                sys.exit(-1)
            else:
                log.info(
                    "A prior Endpoint instance appears to have been terminated without "
                    "proper cleanup. Cleaning up now."
                )
                self.pidfile_cleanup(pid_file)

        result_store = ResultStore(endpoint_dir=endpoint_dir)

        # Create a daemon context
        # If we are running a full detached daemon then we will send the output to
        # log files, otherwise we can piggy back on our stdout
        if endpoint_config.config.detach_endpoint:
            stdout: typing.TextIO = open(
                os.path.join(endpoint_dir, endpoint_config.config.stdout), "a+"
            )
            stderr: typing.TextIO = open(
                os.path.join(endpoint_dir, endpoint_config.config.stderr), "a+"
            )
        else:
            stdout = sys.stdout
            stderr = sys.stderr

        try:
            context = daemon.DaemonContext(
                working_directory=endpoint_dir,
                umask=0o002,
                pidfile=daemon.pidfile.PIDLockFile(pid_file),
                stdout=stdout,
                stderr=stderr,
                detach_process=endpoint_config.config.detach_endpoint,
            )

        except Exception:
            log.exception(
                "Caught exception while trying to setup endpoint context dirs"
            )
            sys.exit(-1)

        # place registration after everything else so that the endpoint will
        # only be registered if everything else has been set up successfully
        reg_info = None
        try:
            reg_info = register_endpoint(
                funcx_client, endpoint_uuid, endpoint_dir, self.name
            )
        # if the service sends back an error response, it will be a FuncxResponseError
        except FuncxResponseError as e:
            # an example of an error that could conceivably occur here would be
            # if the service could not register this endpoint with the forwarder
            # because the forwarder was unreachable
            if e.http_status_code >= 500:
                log.exception("Caught exception while attempting endpoint registration")
                log.critical(
                    "Endpoint registration will be retried in the new endpoint daemon "
                    "process. The endpoint will not work until it is successfully "
                    "registered."
                )
            else:
                raise e
        # if the service has an unexpected internal error and is unable to send
        # back a FuncxResponseError
        except GlobusAPIError as e:
            if e.http_status >= 500:
                log.exception("Caught exception while attempting endpoint registration")
                log.critical(
                    "Endpoint registration will be retried in the new endpoint daemon "
                    "process. The endpoint will not work until it is successfully "
                    "registered."
                )
            else:
                raise e
        # if the service is unreachable due to a timeout or connection error
        except NetworkError as e:
            # the output of a NetworkError exception is huge and unhelpful, so
            # it seems better to just stringify it here and get a concise error
            log.exception(
                f"Caught exception while attempting endpoint registration: {e}"
            )
            log.critical(
                "funcx-endpoint is unable to reach the funcX service due to a "
                "NetworkError \n"
                "Please make sure that the funcX service address you provided is "
                "reachable \n"
                "and then attempt restarting the endpoint"
            )
            exit(-1)
        except Exception:
            raise

        if reg_info:
            log.info("Launching endpoint daemon process")
        else:
            log.critical("Launching endpoint daemon process with errors noted above")

        # NOTE
        # It's important that this log is emitted before we enter the daemon context
        # because daemonization closes down everything, a log message inside the
        # context won't write the currently configured loggers
        logfile = os.path.join(endpoint_dir, "endpoint.log")
        log.info(
            "Logging will be reconfigured for the daemon. logfile=%s , debug=%s",
            logfile,
            self.debug,
        )

        with context:
            setup_logging(
                logfile=logfile,
                debug=self.debug,
                console_enabled=log_to_console,
                no_color=no_color,
            )

            self.daemon_launch(
                endpoint_uuid,
                endpoint_dir,
                endpoint_config,
                reg_info,
                funcx_client,
                result_store,
            )

    def daemon_launch(
        self,
        endpoint_uuid,
        endpoint_dir,
        endpoint_config,
        reg_info,
        funcx_client: FuncXClient,
        result_store: ResultStore,
    ):
        interchange = EndpointInterchange(
            config=endpoint_config.config,
            reg_info=reg_info,
            endpoint_id=endpoint_uuid,
            endpoint_dir=endpoint_dir,
            endpoint_name=self.name,
            funcx_client=funcx_client,
            result_store=result_store,
            logdir=endpoint_dir,
        )

        interchange.start()

        log.critical("Interchange terminated.")

    def stop_endpoint(self, name):
        self.name = name
        endpoint_dir = os.path.join(self.funcx_dir, self.name)
        pid_file = os.path.join(endpoint_dir, "daemon.pid")
        pid_check = self.check_pidfile(pid_file)

        # The process is active if the PID file exists and the process it points to is
        # a funcx-endpoint
        if pid_check["active"]:
            log.debug(f"{self.name} has a daemon.pid file")
            pid = None
            with open(pid_file) as f:
                pid = int(f.read())
            # Attempt terminating
            try:
                log.debug(f"Signalling process: {pid}")
                # For all the processes, including the deamon and its child process tree
                # Send SIGTERM to the processes
                # Wait for 200ms
                # Send SIGKILL to the processes that are still alive
                parent = psutil.Process(pid)
                processes = parent.children(recursive=True)
                processes.append(parent)
                for p in processes:
                    p.send_signal(signal.SIGTERM)
                # graceful RabbitMQ cleanup takes some time
                terminated, alive = psutil.wait_procs(processes, timeout=10)
                for p in alive:
                    # sometimes a process that was marked as alive before can terminate
                    # before this signal is sent
                    try:
                        p.send_signal(signal.SIGKILL)
                    except psutil.NoSuchProcess:
                        pass
                # Wait to confirm that the pid file disappears
                if not os.path.exists(pid_file):
                    log.info(f"Endpoint <{self.name}> is now stopped")

            except OSError:
                log.warning(f"Endpoint <{self.name}> could not be terminated")
                log.warning(f"Attempting Endpoint <{self.name}> cleanup")
                os.remove(pid_file)
                sys.exit(-1)
        # The process is not active, but the PID file exists and needs to be deleted
        elif pid_check["exists"]:
            self.pidfile_cleanup(pid_file)
        else:
            log.info(f"Endpoint <{self.name}> is not active.")

    def delete_endpoint(self, name):
        self.name = name
        endpoint_dir = os.path.join(self.funcx_dir, self.name)

        if not os.path.exists(endpoint_dir):
            log.warning(f"Endpoint <{self.name}> does not exist")
            sys.exit(-1)

        # stopping the endpoint should handle all of the process cleanup before
        # deletion of the directory
        self.stop_endpoint(self.name)

        shutil.rmtree(endpoint_dir)
        log.info(f"Endpoint <{self.name}> has been deleted.")

    def check_pidfile(self, filepath):
        """Helper function to identify possible dead endpoints

        Returns a record with 'exists' and 'active' fields indicating
        whether the pidfile exists, and whether the process is active if it does exist
        (The endpoint can only start correctly if there is no pidfile)

        Parameters
        ----------
        filepath : str
            Path to the pidfile
        """
        if not os.path.exists(filepath):
            return {"exists": False, "active": False}

        pid = int(open(filepath).read().strip())

        active = False
        try:
            psutil.Process(pid)
        except psutil.NoSuchProcess:
            pass
        else:
            # this is the only case where the endpoint is active. If no process exists,
            # it means the endpoint has been terminated without proper cleanup
            active = True

        return {"exists": True, "active": active}

    def pidfile_cleanup(self, filepath):
        os.remove(filepath)
        log.info(f"Endpoint <{self.name}> has been cleaned up.")

    def get_endpoints(self, status_filter=None):
        """
        Gets a dictionary that contains information about all locally
        known endpoints.

        "status" can be one of:
            ["Running", "Disconnected", "Stopped"]

        Example output:
        {
            "default": {
                 "status": "Running",
                 "id": "123abcde-a393-4456-8de5-123456789abc"
            },
            "my_test_ep": {
                 "status": "Disconnected",
                 "id": "xxxxxxxx-xxxx-1234-abcd-xxxxxxxxxxxx"
            }
        }
        """
        endpoint_dict = {}
        config_files = glob.glob(os.path.join(self.funcx_dir, "*", "config.py"))
        for config_file in config_files:
            endpoint_dir = os.path.dirname(config_file)
            endpoint_name = os.path.basename(endpoint_dir)
            status = "Initialized"
            endpoint_id = None

            endpoint_json = os.path.join(endpoint_dir, "endpoint.json")
            if os.path.exists(endpoint_json):
                with open(endpoint_json) as f:
                    endpoint_info = json.load(f)
                    endpoint_id = endpoint_info["endpoint_id"]
                pid_check = self.check_pidfile(os.path.join(endpoint_dir, "daemon.pid"))
                if pid_check["active"]:
                    status = "Running"
                elif pid_check["exists"]:
                    status = "Disconnected"
                else:
                    status = "Stopped"

            if status_filter is None or status_filter == status:
                endpoint_dict[endpoint_name] = {
                    "status": status,
                    "id": endpoint_id,
                }
        return endpoint_dict

    def get_running_endpoints(self):
        return self.get_endpoints(status_filter="Running")

    def print_endpoint_table(self):
        """
        Converts locally configured endpoint list to a text based table
        and prints the output.
          For example format, see the texttable module
        """
        endpoints = self.get_endpoints()
        table = texttable.Texttable()
        headings = ["Endpoint Name", "Status", "Endpoint ID"]
        table.header(headings)

        for endpoint_name, endpoint_info in endpoints.items():
            table.add_row([endpoint_name, endpoint_info["status"], endpoint_info["id"]])
        print(table.draw())



from __future__ import annotations

import logging
import os
import pathlib
from importlib.machinery import SourceFileLoader

import click
from click import ClickException

from funcx.sdk.login_manager import LoginManager

from .endpoint.endpoint import Endpoint
from .logging_config import setup_logging

log = logging.getLogger(__name__)


class CommandState:
    def __init__(self):
        self.endpoint_config_dir: str = str(pathlib.Path.home() / ".funcx")
        self.debug = False
        self.no_color = False
        self.log_to_console = False

    @classmethod
    def ensure(cls) -> CommandState:
        return click.get_current_context().ensure_object(CommandState)


def get_cli_endpoint() -> Endpoint:
    # this getter creates an Endpoint object from the CommandState
    # it takes its various configurable values from the current CommandState
    # as a result, any number of CLI options may be used to tweak the CommandState
    # via callbacks, and the Endpoint will only be constructed within commands which
    # access the Endpoint via this getter
    state = CommandState.ensure()

    endpoint = Endpoint(funcx_dir=state.endpoint_config_dir, debug=state.debug)

    # ensure that configs exist
    if not os.path.exists(endpoint.funcx_dir):
        log.info(
            "No existing configuration found at %s. Initializing...", endpoint.funcx_dir
        )
        endpoint.init_endpoint()

    return endpoint


def log_flag_callback(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return

    state = CommandState.ensure()
    setattr(state, param.name, value)
    setup_logging(debug=state.debug, no_color=state.no_color)


def common_options(f):
    f = click.option(
        "--debug",
        is_flag=True,
        expose_value=False,
        callback=log_flag_callback,
        is_eager=True,
        help="Emit extra information generally only helpful for developers",
    )(f)
    f = click.option(
        "--log-to-console",
        is_flag=True,
        expose_value=False,
        is_eager=True,
        callback=log_flag_callback,
        help="Emit log lines to console as well as log",
    )(f)
    f = click.option(
        "--no-color",
        is_flag=True,
        expose_value=False,
        is_eager=True,
        callback=log_flag_callback,
        help="Colorize the (console) log lines",
    )(f)

    f = click.help_option("-h", "--help")(f)
    return f


def name_arg(f):
    return click.argument("name", required=False, default="default")(f)


def config_dir_callback(ctx, param, value):
    if value is None or ctx.resilient_parsing:
        return
    state = CommandState.ensure()
    state.endpoint_config_dir = value


@click.group("funcx-endpoint")
@click.option(
    "-c",
    "--config-dir",
    default=None,
    help="override default config dir",
    callback=config_dir_callback,
    expose_value=False,
)
def app():
    # the main command group body runs on every command, so the block below will always
    # execute
    setup_logging()  # Until we parse the CLI flags, just setup default logging


@app.command("version")
@common_options
def version_command():
    """Show the version of funcx-endpoint"""
    import funcx_endpoint

    click.echo(f"FuncX endpoint version: {funcx_endpoint.__version__}")


@app.command(name="configure", help="Configure an endpoint")
@click.option("--endpoint-config", default=None, help="a template config to use")
@name_arg
@common_options
def configure_endpoint(*, name: str, endpoint_config: str | None):
    """Configure an endpoint

    Drops a config.py template into the funcx configs directory.
    The template usually goes to ~/.funcx/<ENDPOINT_NAME>/config.py
    """
    endpoint = get_cli_endpoint()
    endpoint.configure_endpoint(name, endpoint_config)


@app.command(name="start", help="Start an endpoint by name")
@name_arg
@click.option(
    "--endpoint-uuid", default=None, help="The UUID for the endpoint to register with"
)
@common_options
def start_endpoint(*, name: str, endpoint_uuid: str | None):
    """Start an endpoint

    This function will do:
    1. Connect to the broker service, and register itself
    2. Get connection info from broker service
    3. Start the interchange as a daemon


    |                      Broker service       |
    |               -----2----> Forwarder       |
    |    /register <-----3----+   ^             |
    +-----^-----------------------+-------------+
          |     |                 |
          1     4                 6
          |     v                 |
    +-----+-----+-----+           v
    |      Start      |---5---> EndpointInterchange
    |     Endpoint    |         daemon
    +-----------------+
    """
    state = CommandState.ensure()
    _do_start_endpoint(
        name=name,
        endpoint_uuid=endpoint_uuid,
        log_to_console=state.log_to_console,
        no_color=state.no_color,
    )


@app.command(name="logout", help="Logout from all endpoints")
@click.option(
    "--force",
    is_flag=True,
    help="Revokes tokens even with currently running endpoints",
)
def logout_endpoints(force: bool):
    success, msg = _do_logout_endpoints(force=force)
    if not success:
        # Raising ClickException is apparently the way to do sys.exit(1)
        #     and return a non-zero value to the command line
        # See https://click.palletsprojects.com/en/8.1.x/exceptions/
        if not isinstance(msg, str) or msg is None:
            # Generic unsuccessful if no reason was given
            msg = "Logout unsuccessful"
        raise ClickException(msg)


def _do_logout_endpoints(
    force: bool, running_endpoints: dict | None = None
) -> tuple[bool, str | None]:
    """
    Logout from all endpoints and remove cached authentication credentials

    Returns True, None if logout was successful and tokens were found and revoked
    Returns False, error_msg if token revocation was not done
    """

    if running_endpoints is None:
        running_endpoints = get_cli_endpoint().get_running_endpoints()
    tokens_revoked = False
    error_msg = None
    if running_endpoints and not force:
        running_list = ", ".join(running_endpoints.keys())
        log.info(
            "The following endpoints are currently running: "
            + running_list
            + "\nPlease use logout --force to proceed"
        )
        error_msg = "Not logging out with running endpoints without --force"
    else:
        tokens_revoked = LoginManager().logout()
        if tokens_revoked:
            log.info("Logout succeeded and all cached credentials were revoked")
        else:
            error_msg = "No cached tokens were found, already logged out?"
            log.info(error_msg)
    return tokens_revoked, error_msg


def _do_start_endpoint(
    *,
    name: str,
    endpoint_uuid: str | None,
    log_to_console: bool,
    no_color: bool,
):
    endpoint = get_cli_endpoint()
    endpoint_dir = os.path.join(endpoint.funcx_dir, name)

    if not os.path.exists(endpoint_dir):
        configure_command = "funcx-endpoint configure"
        if name != "default":
            configure_command += f" -n {name}"
        msg = (
            f"\nEndpoint {name} is not configured!\n"
            "1. Please create a configuration template with:\n"
            f"\t{configure_command}\n"
            "2. Update the configuration\n"
            "3. Start the endpoint\n"
        )
        click.echo(msg)
        return

    try:
        endpoint_config = SourceFileLoader(
            "config", os.path.join(endpoint_dir, "config.py")
        ).load_module()
    except Exception:
        log.exception(
            "funcX v0.2.0 made several non-backwards compatible changes to the config. "
            "Your config might be out of date. "
            "Refer to "
            "https://funcx.readthedocs.io/en/latest/endpoints.html#configuring-funcx"
        )
        raise
    endpoint.start_endpoint(
        name, endpoint_uuid, endpoint_config, log_to_console, no_color
    )


@app.command("stop")
@name_arg
@common_options
def stop_endpoint(*, name: str):
    """Stops an endpoint using the pidfile"""
    _do_stop_endpoint(name=name)


def _do_stop_endpoint(*, name: str) -> None:
    endpoint = get_cli_endpoint()
    endpoint.stop_endpoint(name)


@app.command("restart")
@name_arg
@click.option("--endpoint-uuid", default=None, hidden=True)
@common_options
def restart_endpoint(*, name: str, endpoint_uuid: str | None):
    """Restarts an endpoint"""
    state = CommandState.ensure()
    _do_stop_endpoint(name=name)
    _do_start_endpoint(
        name=name,
        endpoint_uuid=endpoint_uuid,
        log_to_console=state.log_to_console,
        no_color=state.no_color,
    )


@app.command("list")
@common_options
def list_endpoints():
    """List all available endpoints"""
    endpoint = get_cli_endpoint()
    endpoint.print_endpoint_table()


@app.command("delete")
@name_arg
@click.option("--yes", is_flag=True, help="Do not ask for confirmation to delete.")
@common_options
def delete_endpoint(*, name: str, yes: bool):
    """Deletes an endpoint and its config."""
    if not yes:
        click.confirm(
            f"Are you sure you want to delete the endpoint <{name}>?", abort=True
        )

    endpoint = get_cli_endpoint()
    endpoint.delete_endpoint(name)


def cli_run():
    """Entry point for setuptools to point to"""
    app()


if __name__ == "__main__":
    app()



from __future__ import annotations

import logging
import pickle
import uuid

from funcx_common.messagepack import InvalidMessageError
from funcx_common.messagepack import Message as OutgoingMessage
from funcx_common.messagepack import pack, unpack
from funcx_common.messagepack.message_types import (
    EPStatusReport as OutgoingEPStatusReport,
)
from funcx_common.messagepack.message_types import Result as OutgoingResult
from funcx_common.messagepack.message_types import (
    ResultErrorDetails as OutgoingResultErrorDetails,
)
from funcx_common.messagepack.message_types import Task as OutgoingTask

from funcx_endpoint.executors.high_throughput.messages import (
    EPStatusReport as InternalEPStatusReport,
)
from funcx_endpoint.executors.high_throughput.messages import Message as InternalMessage
from funcx_endpoint.executors.high_throughput.messages import Task as InternalTask

logger = logging.getLogger(__name__)


def try_convert_to_messagepack(message: bytes) -> bytes:
    try:
        unpacked = pickle.loads(message)
    except pickle.UnpicklingError:
        # message isn't pickled; assume that it's already in messagepack format
        return message

    messagepack_msg: OutgoingMessage | None = None

    if isinstance(unpacked, InternalEPStatusReport):
        messagepack_msg = OutgoingEPStatusReport(
            endpoint_id=unpacked._header,
            ep_status_report=unpacked.ep_status,
            task_statuses=unpacked.task_statuses,
        )
    elif isinstance(unpacked, dict):
        # exec_start|end_ms is not currently being used in prod and also
        # may not be present
        kwargs = {
            "task_id": uuid.UUID(unpacked["task_id"]),
            "exec_start_ms": unpacked.get("exec_start_ms", 0),
            "exec_end_ms": unpacked.get("exec_end_ms", 0),
        }
        if "exception" in unpacked:
            kwargs["data"] = unpacked["exception"]
            code, user_message = unpacked.get("error_details", ("Unknown", "Unknown"))
            kwargs["error_details"] = OutgoingResultErrorDetails(
                code=code, user_message=user_message
            )
        else:
            kwargs["data"] = unpacked["data"]

        messagepack_msg = OutgoingResult(**kwargs)

    if messagepack_msg:
        message = pack(messagepack_msg)

    return message


def try_convert_from_messagepack(message: bytes) -> bytes:
    try:
        unpacked = unpack(message)
    except InvalidMessageError:
        # message isn't in messagepack form,
        # assume it's already in internal message form
        return message

    internal_message: InternalMessage | None = None

    if isinstance(unpacked, OutgoingTask):
        container_id = "RAW" if unpacked.container_id is None else unpacked.container_id
        internal_message = InternalTask(
            task_id=str(unpacked.task_id),
            container_id=str(container_id),
            task_buffer=unpacked.task_buffer,
        )

    if internal_message:
        message = internal_message.pack()

    return message



from .base import RabbitPublisherStatus, SubscriberProcessStatus
from .result_queue_publisher import ResultQueuePublisher
from .task_queue_subscriber import TaskQueueSubscriber

__all__ = (
    "TaskQueueSubscriber",
    "ResultQueuePublisher",
    "RabbitPublisherStatus",
    "SubscriberProcessStatus",
)



import multiprocessing as mp
import platform
import typing as t

mpQueue: t.Type[mp.Queue]

if platform.system() == "Darwin":
    from parsl.multiprocessing import MacSafeQueue as mpQueue
else:
    from multiprocessing import Queue as mpQueue

__all__ = ("mpQueue",)



#!/usr/bin/env python3

import logging
import time

import dill
import zmq

from funcx_endpoint.executors.high_throughput.messages import Message

log = logging.getLogger(__name__)


class CommandClient:
    """CommandClient"""

    def __init__(self, ip_address, port_range):
        """
        Parameters
        ----------

        ip_address: str
           IP address of the client (where Parsl runs)
        port_range: tuple(int, int)
           Port range for the comms between client and interchange

        """
        self.context = zmq.Context()
        self.zmq_socket = self.context.socket(zmq.DEALER)
        self.zmq_socket.set_hwm(0)
        self.port = self.zmq_socket.bind_to_random_port(
            f"tcp://{ip_address}",
            min_port=port_range[0],
            max_port=port_range[1],
        )

    def run(self, message):
        """This function needs to be fast at the same time aware of the possibility of
        ZMQ pipes overflowing.

        The timeout increases slowly if contention is detected on ZMQ pipes.
        We could set copy=False and get slightly better latency but this results
        in ZMQ sockets reaching a broken state once there are ~10k tasks in flight.
        This issue can be magnified if each the serialized buffer itself is larger.
        """
        self.zmq_socket.send(message.pack(), copy=True)
        reply = self.zmq_socket.recv()
        return Message.unpack(reply)

    def close(self):
        self.zmq_socket.close()
        self.context.term()


class TasksOutgoing:
    """Outgoing task queue from the executor to the Interchange"""

    def __init__(self, ip_address, port_range):
        """
        Parameters
        ----------

        ip_address: str
           IP address of the client (where Parsl runs)
        port_range: tuple(int, int)
           Port range for the comms between client and interchange

        """
        self.context = zmq.Context()
        self.zmq_socket = self.context.socket(zmq.DEALER)
        self.zmq_socket.set_hwm(0)
        self.port = self.zmq_socket.bind_to_random_port(
            f"tcp://{ip_address}",
            min_port=port_range[0],
            max_port=port_range[1],
        )
        self.poller = zmq.Poller()
        self.poller.register(self.zmq_socket, zmq.POLLOUT)

    def put(self, message, max_timeout=1000):
        """This function needs to be fast at the same time aware of the possibility of
        ZMQ pipes overflowing.

        The timeout increases slowly if contention is detected on ZMQ pipes.
        We could set copy=False and get slightly better latency but this results
        in ZMQ sockets reaching a broken state once there are ~10k tasks in flight.
        This issue can be magnified if each the serialized buffer itself is larger.

        Parameters
        ----------

        message : py object
             Python object to send
        max_timeout : int
             Max timeout in milliseconds that we will wait for before raising an
             exception

        Raises
        ------

        zmq.EAGAIN if the send failed.

        """
        timeout_ms = 0
        current_wait = 0
        while current_wait < max_timeout:
            socks = dict(self.poller.poll(timeout=timeout_ms))
            if self.zmq_socket in socks and socks[self.zmq_socket] == zmq.POLLOUT:
                # The copy option adds latency but reduces the risk of ZMQ overflow
                self.zmq_socket.send(message, copy=True)
                return
            else:
                timeout_ms += 1
                log.debug(
                    "Not sending due to full zmq pipe, timeout: {} ms".format(
                        timeout_ms
                    )
                )
            current_wait += timeout_ms

        # Send has failed.
        log.debug(f"Remote side has been unresponsive for {current_wait}")
        raise zmq.error.Again

    def close(self):
        self.zmq_socket.close()
        self.context.term()


class ResultsIncoming:
    """Incoming results queue from the Interchange to the executor"""

    def __init__(self, ip_address, port_range):
        """
        Parameters
        ----------

        ip_address: str
           IP address of the client (where Parsl runs)
        port_range: tuple(int, int)
           Port range for the comms between client and interchange

        """
        self.context = zmq.Context()
        self.results_receiver = self.context.socket(zmq.DEALER)
        self.results_receiver.set_hwm(0)
        self.port = self.results_receiver.bind_to_random_port(
            f"tcp://{ip_address}",
            min_port=port_range[0],
            max_port=port_range[1],
        )

    def get(self, block=True, timeout=None):
        block_messages = self.results_receiver.recv()
        try:
            res = dill.loads(block_messages)
        except dill.UnpicklingError:
            try:
                res = Message.unpack(block_messages)
            except Exception:
                log.exception(
                    "Message in results queue is not pickle/Message formatted: %s",
                    block_messages,
                )
        return res

    def request_close(self):
        status = self.results_receiver.send(dill.dumps(None))
        time.sleep(0.1)
        return status

    def close(self):
        self.results_receiver.close()
        self.context.term()



import logging
import math
import time

from funcx_endpoint.strategies.base import BaseStrategy

log = logging.getLogger(__name__)


class KubeSimpleStrategy(BaseStrategy):
    """Implements the simple strategy for Kubernetes"""

    def __init__(self, *args, threshold=20, interval=1, max_idletime=60):
        """Initialize the flowcontrol object.

        We start the timer thread here

        Parameters
        ----------
        threshold:(int)
          Tasks after which the callback is triggered

        interval (int)
          seconds after which timer expires

        max_idletime: (int)
          maximum idle time(seconds) allowed for resources after which strategy will
          try to kill them.
          default: 60s

        """
        log.info("KubeSimpleStrategy Initialized")
        super().__init__(*args, threshold=threshold, interval=interval)
        self.max_idletime = max_idletime
        self.executors_idle_since = {}

    def strategize(self, *args, **kwargs):
        try:
            self._strategize(*args, **kwargs)
        except Exception as e:
            log.exception(f"Caught error in strategize : {e}")

    def _strategize(self, *args, **kwargs):
        max_pods = self.interchange.provider.max_blocks
        min_pods = self.interchange.provider.min_blocks

        # Kubernetes provider only supports one manager in a pod
        managers_per_pod = 1

        workers_per_pod = self.interchange.max_workers_per_node
        if workers_per_pod == float("inf"):
            workers_per_pod = 1

        parallelism = self.interchange.provider.parallelism

        active_tasks = self.interchange.get_total_tasks_outstanding()
        log.trace(f"Pending tasks : {active_tasks}")

        status = self.interchange.provider_status()
        log.trace(f"Provider status : {status}")

        for task_type in active_tasks.keys():
            active_pods = status.get(task_type, 0)
            active_slots = active_pods * workers_per_pod * managers_per_pod
            active_tasks_per_type = active_tasks[task_type]

            log.debug(
                "Endpoint has %s active tasks of %s, %s active blocks, "
                "%s connected workers for %s",
                active_tasks_per_type,
                task_type,
                active_pods,
                self.interchange.get_total_live_workers(),
                task_type,
            )

            # Reset the idle time if we are currently running tasks
            if active_tasks_per_type > 0:
                self.executors_idle_since[task_type] = None

            # Scale down only if there are no active tasks to avoid having to find which
            # workers are unoccupied
            if active_tasks_per_type == 0 and active_pods > min_pods:
                # We want to make sure that max_idletime is reached before killing off
                # resources
                if not self.executors_idle_since[task_type]:
                    log.debug(
                        "Endpoint has 0 active tasks of task type %s; "
                        "starting kill timer (if idle time exceeds %s seconds, "
                        "resources will be removed)",
                        task_type,
                        self.max_idletime,
                    )
                    self.executors_idle_since[task_type] = time.time()

                # If we have resources idle for the max duration we have to scale_in now
                if (
                    time.time() - self.executors_idle_since[task_type]
                ) > self.max_idletime:
                    log.info(
                        "Idle time has reached %s seconds; "
                        "removing resources of task type %s",
                        self.max_idletime,
                        task_type,
                    )
                    self.interchange.scale_in(
                        active_pods - min_pods, task_type=task_type
                    )
            # More tasks than the available slots.
            elif (
                active_tasks_per_type > 0
                and (float(active_slots) / active_tasks_per_type) < parallelism
            ):
                if active_pods < max_pods:
                    excess = math.ceil(
                        (active_tasks_per_type * parallelism) - active_slots
                    )
                    excess_blocks = math.ceil(
                        float(excess) / (workers_per_pod * managers_per_pod)
                    )
                    excess_blocks = min(excess_blocks, max_pods - active_pods)
                    log.info(f"Requesting {excess_blocks} more blocks")
                    self.interchange.scale_out(excess_blocks, task_type=task_type)
            # Immediatly scale if we are stuck with zero pods and work to do
            elif active_slots == 0 and active_tasks_per_type > 0:
                log.info("Requesting single pod")
                self.interchange.scale_out(1, task_type=task_type)



from parsl.providers import LocalProvider

from funcx_endpoint.endpoint.utils.config import Config
from funcx_endpoint.executors import HighThroughputExecutor

config = Config(
    executors=[
        HighThroughputExecutor(
            provider=LocalProvider(
                init_blocks=1,
                min_blocks=0,
                max_blocks=1,
            ),
        )
    ],
    funcx_service_address="https://api2.funcx.org/v2",
)

# For now, visible_to must be a list of URNs for globus auth users or groups, e.g.:
# urn:globus:auth:identity:{user_uuid}
# urn:globus:groups:id:{group_uuid}
meta = {
    "name": "$name",
    "description": "",
    "organization": "",
    "department": "",
    "public": False,
    "visible_to": [],
}



import os

from setuptools import find_packages, setup

REQUIRES = [
    "requests>=2.20.0,<3",
    "globus-sdk",  # version will be bounded by `funcx`
    "funcx>=1.0.0",
    "funcx-common==0.0.15",
    # table printing used in list-endpoints
    "texttable>=1.6.4,<2",
    # although psutil does not declare itself to use semver, it appears to offer
    # strong backwards-compatibility promises based on its changelog, usage, and
    # history
    #
    # TODO: re-evaluate bound after we have an answer of some kind from psutil
    # see:
    #   https://github.com/giampaolo/psutil/issues/2002
    "psutil<6",
    # provides easy daemonization of the endpoint
    "python-daemon>=2,<3",
    # CLI parsing
    "click>=8,<9",
    # disallow use of 22.3.0; the whl package on some platforms causes ZMQ issues
    #
    # NOTE: 22.3.0 introduced a patched version of libzmq.so to the wheel packaging
    # which may be the source of the problems , the problem can be fixed by
    # building from source, which may mean there's an issue in the packaged library
    # further investigation may be needed if the issue persists in the next pyzmq
    # release
    "pyzmq>=22.0.0,!=22.3.0,<=23.2.0",
    # TODO: evaluate removal of the 'retry' library after the update to
    # globus-sdk v3, which provides automatic retries on all API calls
    "retry==0.9.2",
    # 'parsl' is a core requirement of the funcx-endpoint, essential to a range
    # of different features and functions
    # pin exact versions because it does not use semver
    "parsl>=2022.10.4",
    "pika>=1.2.0",
]

TEST_REQUIRES = [
    "responses",
    "pytest>=5.2",
    "coverage>=5.2",
    "pytest-mock==3.2.0",
    "pyfakefs",
]


version_ns = {}
with open(os.path.join("funcx_endpoint", "version.py")) as f:
    exec(f.read(), version_ns)
version = version_ns["__version__"]

setup(
    name="funcx-endpoint",
    version=version,
    packages=find_packages(),
    description="funcX: High Performance Function Serving for Science",
    install_requires=REQUIRES,
    extras_require={
        "test": TEST_REQUIRES,
    },
    python_requires=">=3.7",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering",
    ],
    keywords=["funcX", "FaaS", "Function Serving"],
    entry_points={
        "console_scripts": [
            "funcx-endpoint=funcx_endpoint.cli:cli_run",
            "funcx-interchange"
            "=funcx_endpoint.executors.high_throughput.interchange:cli_run",
            "funcx-manager"
            "=funcx_endpoint.executors.high_throughput.funcx_manager:cli_run",
            "funcx-worker"
            "=funcx_endpoint.executors.high_throughput.funcx_worker:cli_run",
        ]
    },
    include_package_data=True,
    author="funcX team",
    author_email="labs@globus.org",
    license="Apache License, Version 2.0",
    url="https://github.com/funcx-faas/funcx",
)



import json
import logging
import os
import re
import typing as t

import funcx_endpoint
from funcx import FuncXClient

log = logging.getLogger(__name__)


def register_endpoint(
    funcx_client: FuncXClient, endpoint_uuid: str, endpoint_dir: str, endpoint_name: str
) -> t.Tuple[t.Dict, t.Dict]:
    """Register the endpoint and return the registration info. This function needs
    to be isolated so that the function can both be called from the endpoint start
    process as well as the daemon process that it spawns.

    Parameters
    ----------
    funcx_client : FuncXClient
        The auth'd client to communicate with the funcX service

    endpoint_uuid : str
        The uuid to register the endpoint with

    endpoint_dir : str
        The endpoint directory path to store data in

    endpoint_name : str
        The name of the endpoint

    Returns: tuple[dict, dict]
        Eg return value:
        ({'pika_conn_params': <PIKA_URLParameters>,
          'exchange_name': 'tasks',
          'exchange_type': 'direct',
          'task_url': 'amqp://funcx:rabbitmq@192.168.49.2:5672/'},
         {'pika_conn_params': <PIKA_URLParameters>,
          'exchange_name': 'results',
          'exchange_type': 'topic',
          'task_url': 'amqp://funcx:rabbitmq@192.168.49.2:5672/'}
         )
    """
    log.debug("Attempting registration")
    log.debug(f"Trying with eid : {endpoint_uuid}")

    reg_info = funcx_client.register_endpoint(
        endpoint_name,
        endpoint_uuid,
        endpoint_version=funcx_endpoint.__version__,
    )
    if reg_info.get("endpoint_id") != endpoint_uuid:
        raise ValueError("Unexpected response from server: mismatched endpoint id.")

    # sanitize passwords in logs
    log_reg_info = re.subn(r"://.*?@", r"://***:***@", repr(reg_info))
    log.info(f"Registration returned: {log_reg_info}")

    # NOTE: While all registration info is saved to endpoint.json, only the
    # endpoint UUID is reused from this file.
    with open(os.path.join(endpoint_dir, "endpoint.json"), "w+") as fp:
        endpoint_info = {
            "endpoint_name": endpoint_name,
            # This is named endpoint_id for backward compatibility when
            # funcx-endpoint list is called
            "endpoint_id": endpoint_uuid,
        }
        json.dump(endpoint_info, fp)
        log.debug(
            "Registration info written to {}".format(
                os.path.join(endpoint_dir, "endpoint.json")
            )
        )

    return reg_info["task_queue_info"], reg_info["result_queue_info"]



import os

import globus_sdk
from parsl.addresses import address_by_route
from parsl.channels import LocalChannel
from parsl.config import Config
from parsl.executors import HighThroughputExecutor
from parsl.providers import KubernetesProvider, LocalProvider

# GlobusAuth-related secrets
SECRET_KEY = os.environ.get("secret_key")
GLOBUS_KEY = os.environ.get("globus_key")
GLOBUS_CLIENT = os.environ.get("globus_client")

FUNCX_URL = "https://funcx.org/"
FUNCX_HUB_URL = "3.88.81.131"


def _load_auth_client():
    """Create an AuthClient for the portal

    No credentials are used if the server is not production

    Returns
    -------
    globus_sdk.ConfidentialAppAuthClient
        Client used to perform GlobusAuth actions
    """

    _prod = True

    if _prod:
        app = globus_sdk.ConfidentialAppAuthClient(GLOBUS_CLIENT, GLOBUS_KEY)
    else:
        app = globus_sdk.ConfidentialAppAuthClient("", "")
    return app


def _get_parsl_config():
    """Get the Parsl config.

    Returns
    -------
    parsl.config.Config
        Parsl config to execute tasks.
    """

    config = Config(
        executors=[
            HighThroughputExecutor(
                label="htex_local",
                worker_debug=False,
                cores_per_worker=1,
                max_workers=1,
                provider=LocalProvider(
                    channel=LocalChannel(),
                    init_blocks=1,
                    max_blocks=1,
                    min_blocks=1,
                ),
            )
        ],
        strategy=None,
    )
    return config


def _get_executor(container):
    """
    Get the Parsl executor from the container

    Returns
    -----------
    executor
    """

    executor = HighThroughputExecutor(
        label=container["container_uuid"],
        cores_per_worker=1,
        max_workers=1,
        poll_period=10,
        # launch_cmd="ls; sleep 3600",
        worker_logdir_root="runinfo",
        # worker_debug=True,
        address=address_by_route(),
        provider=KubernetesProvider(
            namespace="dlhub-privileged",
            image=container["location"],
            nodes_per_block=1,
            init_blocks=1,
            max_blocks=1,
            parallelism=1,
            worker_init="""pip install git+https://github.com/Parsl/parsl;
                                   pip install git+https://github.com/funcx-faas/funcX;
                                   export PYTHONPATH=$PYTHONPATH:/home/ubuntu:/app""",
            # security=None,
            secret="ryan-kube-secret",
            pod_name=container["name"]
            .replace(".", "-")
            .replace("_", "-")
            .replace("/", "-")
            .lower(),
            # secret="minikube-aws-ecr",
            # user_id=32781,
            # group_id=10253,
            # run_as_non_root=True
        ),
    )
    return [executor]



from __future__ import annotations

import argparse
import logging
import os
import signal
import sys
import time

import dill
import zmq
from funcx_common import messagepack

from funcx.errors import MaxResultSizeExceeded
from funcx.serialize import FuncXSerializer
from funcx_endpoint.exception_handling import get_error_string, get_result_error_details
from funcx_endpoint.exceptions import CouldNotExecuteUserTaskError
from funcx_endpoint.executors.high_throughput.messages import Message
from funcx_endpoint.logging_config import setup_logging

log = logging.getLogger(__name__)

DEFAULT_RESULT_SIZE_LIMIT_MB = 10
DEFAULT_RESULT_SIZE_LIMIT_B = DEFAULT_RESULT_SIZE_LIMIT_MB * 1024 * 1024


def _now_ms() -> int:
    return int(time.time() * 1000)


class FuncXWorker:
    """The FuncX worker
    Parameters
    ----------

    worker_id : str
     Worker id string

    address : str
     Address at which the manager might be reached. This is usually 127.0.0.1

    port : int
     Port at which the manager can be reached

    result_size_limit : int
     Maximum result size allowed in Bytes
     Default = 10 MB

    Funcx worker will use the REP sockets to:
         task = recv ()
         result = execute(task)
         send(result)
    """

    def __init__(
        self,
        worker_id,
        address,
        port,
        worker_type="RAW",
        result_size_limit=DEFAULT_RESULT_SIZE_LIMIT_B,
    ):

        self.worker_id = worker_id
        self.address = address
        self.port = port
        self.worker_type = worker_type
        self.serializer = FuncXSerializer()
        self.serialize = self.serializer.serialize
        self.deserialize = self.serializer.deserialize
        self.result_size_limit = result_size_limit

        log.info(f"Initializing worker {worker_id}")
        log.info(f"Worker is of type: {worker_type}")

        self.context = zmq.Context()
        self.poller = zmq.Poller()
        self.identity = worker_id.encode()

        self.task_socket = self.context.socket(zmq.DEALER)
        self.task_socket.setsockopt(zmq.IDENTITY, self.identity)

        log.info(f"Trying to connect to : tcp://{self.address}:{self.port}")
        self.task_socket.connect(f"tcp://{self.address}:{self.port}")
        self.poller.register(self.task_socket, zmq.POLLIN)
        signal.signal(signal.SIGTERM, self.handler)

    def handler(self, signum, frame):
        log.error(f"Signal handler called with signal {signum}")
        sys.exit(1)

    def _send_registration_message(self):
        log.debug("Sending registration")
        payload = {"worker_id": self.worker_id, "worker_type": self.worker_type}
        self.task_socket.send_multipart([b"REGISTER", dill.dumps(payload)])

    def start(self):
        log.info("Starting worker")
        self._send_registration_message()

        while True:
            log.debug("Waiting for task")
            p_task_id, p_container_id, msg = self.task_socket.recv_multipart()
            task_id: str = dill.loads(p_task_id)
            container_id: str = dill.loads(p_container_id)
            log.debug(f"Received task with task_id='{task_id}' and msg='{msg}'")

            result = None
            if task_id == "KILL":
                log.info("[KILL] -- Worker KILL message received! ")
                # send a "worker die" message back to the manager
                self.task_socket.send_multipart([b"WRKR_DIE", b""])
                log.info(f"*** WORKER {self.worker_id} ABOUT TO DIE ***")
                # Kill the worker after accepting death in message to manager.
                sys.exit()
            else:
                result = self.execute_task(task_id, msg)
                result["container_id"] = container_id
                log.debug("Sending result")
                # send bytes over the socket back to the manager
                self.task_socket.send_multipart([b"TASK_RET", dill.dumps(result)])

        log.warning("Broke out of the loop... dying")

    def execute_task(self, task_id: str, task_body: bytes) -> dict:
        log.debug("executing task task_id='%s'", task_id)
        exec_start_ms = _now_ms()

        try:
            result = self.call_user_function(task_body)
        except Exception:
            log.exception("Caught an exception while executing user function")
            result_message = dict(
                task_id=task_id,
                exception=get_error_string(),
                error_details=get_result_error_details(),
                exec_start_ms=exec_start_ms,
                exec_end_ms=_now_ms(),
            )
        else:
            log.debug("Execution completed without exception")
            result_message = dict(
                task_id=task_id,
                data=result,
                exec_start_ms=exec_start_ms,
                exec_end_ms=_now_ms(),
            )

        log.debug("task %s completed", task_id)
        return result_message

    def call_user_function(self, message: bytes) -> str:
        """Deserialize the buffer and execute the task.

        Returns the result or throws exception.
        """
        # try to unpack it as a messagepack message
        try:
            task = messagepack.unpack(message)
            if not isinstance(task, messagepack.message_types.Task):
                raise CouldNotExecuteUserTaskError(
                    f"wrong type of message in worker: {type(task)}"
                )
            task_data = task.task_buffer
        # on parse errors, failover to trying the "legacy" message reading
        except (
            messagepack.InvalidMessageError,
            messagepack.UnrecognizedProtocolVersion,
        ):
            task = Message.unpack(message)
            task_data = task.task_buffer.decode("utf-8")  # type: ignore[attr-defined]

        f, args, kwargs = self.serializer.unpack_and_deserialize(task_data)
        result_data = f(*args, **kwargs)
        serialized_data = self.serialize(result_data)

        if len(serialized_data) > self.result_size_limit:
            raise MaxResultSizeExceeded(len(serialized_data), self.result_size_limit)

        return serialized_data


def cli_run():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-w", "--worker_id", required=True, help="ID of worker from process_worker_pool"
    )
    parser.add_argument(
        "-t", "--type", required=False, help="Container type of worker", default="RAW"
    )
    parser.add_argument(
        "-a", "--address", required=True, help="Address for the manager, eg X,Y,"
    )
    parser.add_argument(
        "-p",
        "--port",
        required=True,
        help="Internal port at which the worker connects to the manager",
    )
    parser.add_argument(
        "--logdir", required=True, help="Directory path where worker log files written"
    )
    parser.add_argument(
        "-d",
        "--debug",
        action="store_true",
        help="Directory path where worker log files written",
    )
    args = parser.parse_args()

    setup_logging(
        logfile=os.path.join(args.logdir, f"funcx_worker_{args.worker_id}.log"),
        debug=args.debug,
    )

    # Redirect the stdout and stderr
    stdout_path = os.path.join(args.logdir, f"funcx_worker_{args.worker_id}.stdout")
    stderr_path = os.path.join(args.logdir, f"funcx_worker_{args.worker_id}.stderr")
    with open(stdout_path, "w") as fo, open(stderr_path, "w") as fe:
        # Redirect the stdout
        old_stdout, old_stderr = sys.stdout, sys.stderr
        sys.stdout = fo
        sys.stderr = fe

        try:
            worker = FuncXWorker(
                args.worker_id,
                args.address,
                int(args.port),
                worker_type=args.type,
            )
            worker.start()
        finally:
            # Switch them back
            sys.stdout = old_stdout
            sys.stderr = old_stderr


if __name__ == "__main__":
    cli_run()



from __future__ import annotations

import logging
import math
import time

from parsl.providers.provider_base import JobState

from funcx_endpoint.strategies.base import BaseStrategy

log = logging.getLogger(__name__)


class SimpleStrategy(BaseStrategy):
    """Implements the simple strategy"""

    def __init__(self, *args, threshold=20, interval=1, max_idletime=60):
        """Initialize the flowcontrol object.

        We start the timer thread here

        Parameters
        ----------
        threshold:(int)
          Tasks after which the callback is triggered

        interval (int)
          seconds after which timer expires

        max_idletime: (int)
          maximum idle time(seconds) allowed for resources after which strategy will
          try to kill them.
          default: 60s

        """
        log.info("SimpleStrategy Initialized")
        super().__init__(*args, threshold=threshold, interval=interval)
        self.max_idletime = max_idletime
        self.executors = {"idle_since": None}

        # caching vars; for log noise reduction
        self._prev_info = None
        self._prev_task_breakdown = None
        self._prev_status = None

    def strategize(self, *args, **kwargs):
        try:
            self._strategize(*args, **kwargs)
        except Exception as e:
            log.exception(f"Caught error in strategize : {e}")
            pass

    def _strategize(self, *args, **kwargs):
        task_breakdown = self.interchange.get_outstanding_breakdown()
        if task_breakdown != self._prev_task_breakdown:
            self._prev_task_breakdown = task_breakdown
            log.debug(f"Task breakdown {task_breakdown}")

        min_blocks = self.interchange.provider.min_blocks
        max_blocks = self.interchange.provider.max_blocks

        # Here we assume that each node has atleast 4 workers

        tasks_per_node = self.interchange.max_workers_per_node
        if self.interchange.max_workers_per_node == float("inf"):
            tasks_per_node = 1

        nodes_per_block = self.interchange.provider.nodes_per_block
        parallelism = self.interchange.provider.parallelism

        active_tasks = sum(self.interchange.get_total_tasks_outstanding().values())
        status = self.interchange.provider_status()

        running = sum(1 for x in status if x.state == JobState.RUNNING)
        pending = sum(1 for x in status if x.state == JobState.PENDING)
        active_blocks = running + pending
        active_slots = active_blocks * tasks_per_node * nodes_per_block

        status = str(status)
        if status != self._prev_status:
            self._prev_status = status
            log.debug(f"Provider status: {status}")

        cur_info = (
            active_tasks,
            running,
            pending,
            self.interchange.get_total_live_workers(),
        )
        if cur_info != self._prev_info:
            self._prev_info = cur_info
            log.debug(
                "Endpoint has %s active tasks, %s/%s running/pending blocks, "
                "and %s connected workers",
                *cur_info,
            )

        # reset kill timer if executor has active tasks
        if active_tasks > 0 and self.executors["idle_since"]:
            self.executors["idle_since"] = None

        # Case 1
        # No tasks.
        if active_tasks == 0:
            # Case 1a
            # Fewer blocks that min_blocks
            if active_blocks <= min_blocks:
                # Ignore
                # log.debug("Strategy: Case.1a")
                pass

            # Case 1b
            # More blocks than min_blocks. Scale down
            else:
                # We want to make sure that max_idletime is reached
                # before killing off resources
                if not self.executors["idle_since"]:
                    log.debug(
                        "Endpoint has 0 active tasks; starting kill timer "
                        "(if idle time exceeds %s seconds, resources will be removed)",
                        self.max_idletime,
                    )
                    self.executors["idle_since"] = time.time()

                idle_since = self.executors["idle_since"]
                if (time.time() - idle_since) > self.max_idletime:
                    # We have resources idle for the max duration,
                    # we have to scale_in now.
                    log.debug(
                        "Idle time has reached {}s; removing resources".format(
                            self.max_idletime
                        )
                    )
                    self.interchange.scale_in(active_blocks - min_blocks)

                else:
                    pass
                    # log.debug(
                    #     "Strategy: Case.1b. Waiting for timer : %s", idle_since
                    # )

        # Case 2
        # More tasks than the available slots.
        elif (float(active_slots) / active_tasks) < parallelism:
            # Case 2a
            # We have the max blocks possible
            if active_blocks >= max_blocks:
                # Ignore since we already have the max nodes
                # log.debug("Strategy: Case.2a")
                pass

            # Case 2b
            else:
                # log.debug("Strategy: Case.2b")
                excess = math.ceil((active_tasks * parallelism) - active_slots)
                excess_blocks = math.ceil(
                    float(excess) / (tasks_per_node * nodes_per_block)
                )
                excess_blocks = min(excess_blocks, max_blocks - active_blocks)
                log.debug(f"Requesting {excess_blocks} more blocks")
                self.interchange.scale_out(excess_blocks)

        elif active_slots == 0 and active_tasks > 0:
            # Case 4
            # Check if slots are being lost quickly ?
            log.debug("Requesting single slot")
            if active_blocks < max_blocks:
                self.interchange.scale_out(1)
        # Case 3
        # tasks ~ slots
        else:
            # log.debug("Strategy: Case 3")
            pass
